import contextlib
import io

import numpy as np
import pytest

import asdf
from asdf import generic_io, util


def test_not_set():
    assert util.NotSet is not None

    assert repr(util.NotSet) == "NotSet"


class SomeClass:
    class SomeInnerClass:
        pass


def test_get_class_name():
    assert util.get_class_name(SomeClass()) == "asdf._tests.test_util.SomeClass"
    assert util.get_class_name(SomeClass, instance=False) == "asdf._tests.test_util.SomeClass"
    assert util.get_class_name(SomeClass.SomeInnerClass()) == "asdf._tests.test_util.SomeClass.SomeInnerClass"
    assert (
        util.get_class_name(SomeClass.SomeInnerClass, instance=False)
        == "asdf._tests.test_util.SomeClass.SomeInnerClass"
    )


def test_patched_urllib_parse():
    assert "asdf" in util._patched_urllib_parse.uses_relative
    assert "asdf" in util._patched_urllib_parse.uses_netloc

    import urllib.parse

    assert urllib.parse is not util._patched_urllib_parse
    assert "asdf" not in urllib.parse.uses_relative
    assert "asdf" not in urllib.parse.uses_netloc


@pytest.mark.parametrize(
    ("pattern", "uri", "result"),
    [
        ("asdf://somewhere.org/tags/foo-1.0", "asdf://somewhere.org/tags/foo-1.0", True),
        ("asdf://somewhere.org/tags/foo-1.0", "asdf://somewhere.org/tags/bar-1.0", False),
        ("asdf://somewhere.org/tags/foo-*", "asdf://somewhere.org/tags/foo-1.0", True),
        ("asdf://somewhere.org/tags/foo-*", "asdf://somewhere.org/tags/bar-1.0", False),
        ("asdf://somewhere.org/tags/foo-*", "asdf://somewhere.org/tags/foo-extras/bar-1.0", False),
        ("asdf://*/tags/foo-*", "asdf://anywhere.org/tags/foo-4.9", True),
        ("asdf://*/tags/foo-*", "asdf://anywhere.org/tags/bar-4.9", False),
        ("asdf://*/tags/foo-*", "asdf://somewhere.org/tags/foo-extras/bar-4.9", False),
        ("asdf://**/*-1.0", "asdf://somewhere.org/tags/foo-1.0", True),
        ("asdf://**/*-1.0", "asdf://somewhere.org/tags/foo-2.0", False),
        ("asdf://**/*-1.0", "asdf://somewhere.org/tags/foo-extras/bar-1.0", True),
        ("asdf://**/*-1.0", "asdf://somewhere.org/tags/foo-extras/bar-2.0", False),
        ("asdf://somewhere.org/tags/foo-*", None, False),
        ("**", None, False),
    ],
)
def test_uri_match(pattern, uri, result):
    assert util.uri_match(pattern, uri) is result


@pytest.mark.parametrize(
    ("content", "expected_type"),
    [
        (b"#ASDF blahblahblah", util.FileType.ASDF),
        (b"SIMPLE = T blah blah blah blah", util.FileType.FITS),
        (b"SIMPLY NOT A FITS FILE", util.FileType.UNKNOWN),
        (b"#ASDQ", util.FileType.UNKNOWN),
    ],
)
def test_get_file_type(content, expected_type):
    fd = generic_io.get_file(io.BytesIO(content))
    assert util.get_file_type(fd) == expected_type
    # Confirm that no content was lost
    assert fd.read() == content

    # We've historically had a problem detecting file type
    # of generic_io.InputStream:
    class OnlyHasAReadMethod:
        def __init__(self, content):
            self._fd = io.BytesIO(content)

        def read(self, size=-1):
            return self._fd.read(size)

    fd = generic_io.get_file(OnlyHasAReadMethod(content))
    assert util.get_file_type(fd) == expected_type
    assert fd.read() == content


@pytest.mark.parametrize("input_type", ["filename", "binary_file", "generic_file"])
@pytest.mark.parametrize("tagged", [True, False])
def test_load_yaml(tmp_path, input_type, tagged):
    fn = tmp_path / "test.asdf"
    asdf.AsdfFile({"a": np.zeros(3)}).write_to(fn)

    if input_type == "filename":
        init = fn
        ctx = contextlib.nullcontext()
    elif input_type == "binary_file":
        init = open(fn, "rb")
        ctx = init
    elif input_type == "generic_file":
        init = generic_io.get_file(fn, "r")
        ctx = init

    with ctx:
        tree = util.load_yaml(init, tagged=tagged)
    if tagged:
        assert isinstance(tree["a"], asdf.tagged.TaggedDict)
    else:
        assert not isinstance(tree["a"], asdf.tagged.TaggedDict)


@pytest.mark.parametrize("tagged", [True, False])
def test_load_yaml_recursion(tmp_path, tagged):
    fn = tmp_path / "test.asdf"
    tree = {}
    tree["d"] = {}
    tree["d"]["d"] = tree["d"]
    tree["l"] = []
    tree["l"].append(tree["l"])
    asdf.AsdfFile(tree).write_to(fn)
    tree = util.load_yaml(fn, tagged=tagged)
    assert tree["d"]["d"] is tree["d"]
    assert tree["l"][0] is tree["l"]


@pytest.mark.parametrize("tagged", [True, False])
def test_load_yaml_recursion_with_tags(tagged):
    contents = b"""#ASDF 1.0.0
#ASDF_STANDARD 1.6.0
%YAML 1.1
%TAG ! tag:stsci.edu:asdf/
--- !core/asdf-1.1.0
o: &id001 !some/tag-1.0.0
  inverse: !some/tag-1.0.0
    inverse: *id001
l: &id002 !some/tag-1.0.0
  - *id002
..."""
    tree = util.load_yaml(io.BytesIO(contents), tagged=tagged)
    assert tree["o"] is tree["o"]["inverse"]["inverse"]
    assert tree["l"] is tree["l"][0]



"""
Manages external references in the YAML tree using the `JSON Reference
standard <http://tools.ietf.org/html/draft-pbryan-zyp-json-ref-03>`__
and `JSON Pointer standard <http://tools.ietf.org/html/rfc6901>`__.
"""

import weakref
from collections.abc import Sequence
from contextlib import suppress

import numpy as np

from . import generic_io, treeutil, util
from .util import _patched_urllib_parse

__all__ = ["Reference", "find_references", "make_reference", "resolve_fragment", "resolve_references"]


def resolve_fragment(tree, pointer):
    """
    Resolve a JSON Pointer within the tree.
    """
    pointer = pointer.lstrip("/")
    parts = _patched_urllib_parse.unquote(pointer).split("/") if pointer else []

    for part in parts:
        part_ = part.replace("~1", "/").replace("~0", "~")

        if isinstance(tree, Sequence):
            # Array indexes should be turned into integers
            with suppress(ValueError):
                part_ = int(part_)

        try:
            tree = tree[part_]

        except (TypeError, LookupError) as err:
            msg = f"Unresolvable reference: '{pointer}'"
            raise ValueError(msg) from err

    return tree


class Reference:
    def __init__(self, uri, base_uri=None, asdffile=None, target=None):
        self._uri = uri
        if asdffile is not None:
            self._asdffile = weakref.ref(asdffile)
        self._base_uri = base_uri
        self._target = target

    def _get_target(self, **kwargs):
        if self._target is None:
            base_uri = self._base_uri
            if base_uri is None:
                base_uri = self._asdffile().uri
            uri = generic_io.resolve_uri(base_uri, self._uri)
            asdffile = self._asdffile().open_external(uri, **kwargs)
            parts = _patched_urllib_parse.urlparse(self._uri)
            fragment = parts.fragment
            self._target = resolve_fragment(asdffile.tree, fragment)
        return self._target

    def __repr__(self):
        # repr alone should not force loading of the reference
        if self._target is None:
            return f"<Reference (unloaded) to '{self._uri}'>"

        return f"<Reference to {self._target!r}>"

    def __str__(self):
        # str alone should not force loading of the reference
        if self._target is None:
            return f"<Reference (unloaded) to '{self._uri}'>"

        return str(self._target)

    def __len__(self):
        return len(self._get_target())

    def __getattr__(self, attr):
        if attr == "_tag":
            return None
        try:
            return getattr(self._get_target(), attr)
        except Exception as err:
            msg = f"No attribute '{attr}'"
            raise AttributeError(msg) from err

    def __getitem__(self, item):
        return self._get_target()[item]

    def __setitem__(self, item, val):
        self._get_target()[item] = val

    def __array__(self):
        return np.asarray(self._get_target())

    def __call__(self, **kwargs):
        return self._get_target(**kwargs)

    def __contains__(self, item):
        return item in self._get_target()


def find_references(tree, ctx):
    """
    Find all of the JSON references in the tree, and convert them into
    `Reference` objects.
    """

    def do_find(tree):
        if isinstance(tree, dict) and "$ref" in tree:
            return Reference(tree["$ref"], asdffile=ctx)
        return tree

    return treeutil.walk_and_modify(tree, do_find)


def resolve_references(tree, ctx, **kwargs):
    """
    Resolve all of the references in the tree, by loading the external
    data and inserting it directly into the tree.
    """

    def do_resolve(tree):
        if isinstance(tree, Reference):
            return tree(**kwargs)
        return tree

    tree = find_references(tree, ctx)

    return treeutil.walk_and_modify(tree, do_resolve)


def make_reference(asdffile, path):
    """
    Make a reference to a subtree of the given ASDF file.

    Parameters
    ----------
    asdffile : AsdfFile

    path : list of str and int, optional
        The parts of the path pointing to an item in this tree.
        If omitted, points to the root of the tree.

    Returns
    -------
    reference : reference.Reference
        A reference object.
    """
    path_str = "/".join(x.replace("~", "~0").replace("/", "~1") for x in path)
    target = resolve_fragment(asdffile.tree, path_str)

    if asdffile.uri is None:
        msg = "Can not make a reference to a AsdfFile without an associated URI."
        raise ValueError(msg)
    base_uri = util.get_base_uri(asdffile.uri)
    uri = base_uri + "#" + path_str
    return Reference(uri, target=target)



import warnings

import pytest


def test_warnings_are_errors():
    """
    Smoke test to make sure that warnings cause errors.
    This is here as previously asdf._tests._helpers had
    an ``assert_no_warnings`` function that was removed
    in favor of using a warning filter to turn warnings
    into errors.
    """
    with pytest.raises(UserWarning):
        warnings.warn("test")



from .defragment import defragment
from .diff import diff
from .edit import edit
from .exploded import explode, implode
from .extension import find_extensions
from .info import info
from .search import search
from .tags import list_tags
from .to_yaml import to_yaml
from .validate import validate

__all__ = [
    "defragment",
    "diff",
    "edit",
    "explode",
    "find_extensions",
    "implode",
    "info",
    "list_tags",
    "search",
    "to_yaml",
    "validate",
]



import numpy as np

import asdf
from asdf import AsdfFile
from asdf._commands import main
from asdf._tests._helpers import assert_tree_match


def test_explode_then_implode(tmp_path):
    x = np.arange(0, 10, dtype=float)

    tree = {
        "science_data": x,
        "subset": x[3:-3],
        "skipping": x[::2],
        "not_shared": np.arange(10, 0, -1, dtype=np.uint8),
    }

    path = tmp_path / "original.asdf"
    ff = AsdfFile(tree)
    # Since we're testing with small arrays, force all arrays to be stored
    # in internal blocks rather than letting some of them be automatically put
    # inline.
    ff.write_to(path, all_array_storage="internal")
    with asdf.open(path) as af:
        assert len(af._blocks.blocks) == 2

    result = main.main_from_args(["explode", str(path)])

    assert result == 0

    files = [p.name for p in tmp_path.iterdir()]

    assert "original.asdf" in files
    assert "original_exploded.asdf" in files
    assert "original_exploded0000.asdf" in files
    assert "original_exploded0001.asdf" in files
    assert "original_exploded0002.asdf" not in files

    # compare file sizes of original and exploded files
    original_size = (tmp_path / "original.asdf").stat().st_size
    exploded_size = (tmp_path / "original_exploded.asdf").stat().st_size
    assert original_size > exploded_size

    path = tmp_path / "original_exploded.asdf"
    result = main.main_from_args(["implode", str(path)])

    assert result == 0

    with asdf.open(tmp_path / "original_exploded_all.asdf") as af:
        assert_tree_match(af.tree, tree)
        assert len(af._blocks.blocks) == 2


def test_file_not_found(tmp_path):
    path = tmp_path / "original.asdf"
    assert main.main_from_args(["explode", str(path)]) == 2



from fractions import Fraction
from urllib.parse import urldefrag, urljoin
import re

from asdf._jsonschema._utils import (
    ensure_list,
    equal,
    extras_msg,
    find_additional_properties,
    find_evaluated_item_indexes_by_schema,
    find_evaluated_property_keys_by_schema,
    unbool,
    uniq,
)
from asdf._jsonschema.exceptions import FormatError, ValidationError


def patternProperties(validator, patternProperties, instance, schema):
    if not validator.is_type(instance, "object"):
        return

    for pattern, subschema in patternProperties.items():
        for k, v in instance.items():
            if re.search(pattern, k):
                yield from validator.descend(
                    v, subschema, path=k, schema_path=pattern,
                )


def propertyNames(validator, propertyNames, instance, schema):
    if not validator.is_type(instance, "object"):
        return

    for property in instance:
        yield from validator.descend(instance=property, schema=propertyNames)


def additionalProperties(validator, aP, instance, schema):
    if not validator.is_type(instance, "object"):
        return

    extras = set(find_additional_properties(instance, schema))

    if validator.is_type(aP, "object"):
        for extra in extras:
            yield from validator.descend(instance[extra], aP, path=extra)
    elif not aP and extras:
        if "patternProperties" in schema:
            if len(extras) == 1:
                verb = "does"
            else:
                verb = "do"

            joined = ", ".join(repr(each) for each in sorted(extras))
            patterns = ", ".join(
                repr(each) for each in sorted(schema["patternProperties"])
            )
            error = f"{joined} {verb} not match any of the regexes: {patterns}"
            yield ValidationError(error)
        else:
            error = "Additional properties are not allowed (%s %s unexpected)"
            yield ValidationError(error % extras_msg(extras))


def items(validator, items, instance, schema):
    if not validator.is_type(instance, "array"):
        return

    prefix = len(schema.get("prefixItems", []))
    total = len(instance)
    if items is False and total > prefix:
        message = f"Expected at most {prefix} items, but found {total}"
        yield ValidationError(message)
    else:
        for index in range(prefix, total):
            yield from validator.descend(
                instance=instance[index],
                schema=items,
                path=index,
            )


def additionalItems(validator, aI, instance, schema):
    if (
        not validator.is_type(instance, "array")
        or validator.is_type(schema.get("items", {}), "object")
    ):
        return

    len_items = len(schema.get("items", []))
    if validator.is_type(aI, "object"):
        for index, item in enumerate(instance[len_items:], start=len_items):
            yield from validator.descend(item, aI, path=index)
    elif not aI and len(instance) > len(schema.get("items", [])):
        error = "Additional items are not allowed (%s %s unexpected)"
        yield ValidationError(
            error % extras_msg(instance[len(schema.get("items", [])):]),
        )


def const(validator, const, instance, schema):
    if not equal(instance, const):
        yield ValidationError(f"{const!r} was expected")


def contains(validator, contains, instance, schema):
    if not validator.is_type(instance, "array"):
        return

    matches = 0
    min_contains = schema.get("minContains", 1)
    max_contains = schema.get("maxContains", len(instance))

    for each in instance:
        if validator.evolve(schema=contains).is_valid(each):
            matches += 1
            if matches > max_contains:
                yield ValidationError(
                    "Too many items match the given schema "
                    f"(expected at most {max_contains})",
                    validator="maxContains",
                    validator_value=max_contains,
                )
                return

    if matches < min_contains:
        if not matches:
            yield ValidationError(
                f"{instance!r} does not contain items "
                "matching the given schema",
            )
        else:
            yield ValidationError(
                "Too few items match the given schema (expected at least "
                f"{min_contains} but only {matches} matched)",
                validator="minContains",
                validator_value=min_contains,
            )


def exclusiveMinimum(validator, minimum, instance, schema):
    if not validator.is_type(instance, "number"):
        return

    if instance <= minimum:
        yield ValidationError(
            f"{instance!r} is less than or equal to "
            f"the minimum of {minimum!r}",
        )


def exclusiveMaximum(validator, maximum, instance, schema):
    if not validator.is_type(instance, "number"):
        return

    if instance >= maximum:
        yield ValidationError(
            f"{instance!r} is greater than or equal "
            f"to the maximum of {maximum!r}",
        )


def minimum(validator, minimum, instance, schema):
    if not validator.is_type(instance, "number"):
        return

    if instance < minimum:
        message = f"{instance!r} is less than the minimum of {minimum!r}"
        yield ValidationError(message)


def maximum(validator, maximum, instance, schema):
    if not validator.is_type(instance, "number"):
        return

    if instance > maximum:
        message = f"{instance!r} is greater than the maximum of {maximum!r}"
        yield ValidationError(message)


def multipleOf(validator, dB, instance, schema):
    if not validator.is_type(instance, "number"):
        return

    if isinstance(dB, float):
        quotient = instance / dB
        try:
            failed = int(quotient) != quotient
        except OverflowError:
            # When `instance` is large and `dB` is less than one,
            # quotient can overflow to infinity; and then casting to int
            # raises an error.
            #
            # In this case we fall back to Fraction logic, which is
            # exact and cannot overflow.  The performance is also
            # acceptable: we try the fast all-float option first, and
            # we know that fraction(dB) can have at most a few hundred
            # digits in each part.  The worst-case slowdown is therefore
            # for already-slow enormous integers or Decimals.
            failed = (Fraction(instance) / Fraction(dB)).denominator != 1
    else:
        failed = instance % dB

    if failed:
        yield ValidationError(f"{instance!r} is not a multiple of {dB}")


def minItems(validator, mI, instance, schema):
    if validator.is_type(instance, "array") and len(instance) < mI:
        yield ValidationError(f"{instance!r} is too short")


def maxItems(validator, mI, instance, schema):
    if validator.is_type(instance, "array") and len(instance) > mI:
        yield ValidationError(f"{instance!r} is too long")


def uniqueItems(validator, uI, instance, schema):
    if (
        uI
        and validator.is_type(instance, "array")
        and not uniq(instance)
    ):
        yield ValidationError(f"{instance!r} has non-unique elements")


def pattern(validator, patrn, instance, schema):
    if (
        validator.is_type(instance, "string")
        and not re.search(patrn, instance)
    ):
        yield ValidationError(f"{instance!r} does not match {patrn!r}")


def format(validator, format, instance, schema):
    if validator.format_checker is not None:
        try:
            validator.format_checker.check(instance, format)
        except FormatError as error:
            yield ValidationError(error.message, cause=error.cause)


def minLength(validator, mL, instance, schema):
    if validator.is_type(instance, "string") and len(instance) < mL:
        yield ValidationError(f"{instance!r} is too short")


def maxLength(validator, mL, instance, schema):
    if validator.is_type(instance, "string") and len(instance) > mL:
        yield ValidationError(f"{instance!r} is too long")


def dependentRequired(validator, dependentRequired, instance, schema):
    if not validator.is_type(instance, "object"):
        return

    for property, dependency in dependentRequired.items():
        if property not in instance:
            continue

        for each in dependency:
            if each not in instance:
                message = f"{each!r} is a dependency of {property!r}"
                yield ValidationError(message)


def dependentSchemas(validator, dependentSchemas, instance, schema):
    if not validator.is_type(instance, "object"):
        return

    for property, dependency in dependentSchemas.items():
        if property not in instance:
            continue
        yield from validator.descend(
            instance, dependency, schema_path=property,
        )


def enum(validator, enums, instance, schema):
    if instance == 0 or instance == 1:
        unbooled = unbool(instance)
        if all(unbooled != unbool(each) for each in enums):
            yield ValidationError(f"{instance!r} is not one of {enums!r}")
    elif instance not in enums:
        yield ValidationError(f"{instance!r} is not one of {enums!r}")


def ref(validator, ref, instance, schema):
    resolve = getattr(validator.resolver, "resolve", None)
    if resolve is None:
        with validator.resolver.resolving(ref) as resolved:
            yield from validator.descend(instance, resolved)
    else:
        scope, resolved = validator.resolver.resolve(ref)
        validator.resolver.push_scope(scope)

        try:
            yield from validator.descend(instance, resolved)
        finally:
            validator.resolver.pop_scope()


def dynamicRef(validator, dynamicRef, instance, schema):
    _, fragment = urldefrag(dynamicRef)

    for url in validator.resolver._scopes_stack:
        lookup_url = urljoin(url, dynamicRef)
        with validator.resolver.resolving(lookup_url) as subschema:
            if ("$dynamicAnchor" in subschema
                    and fragment == subschema["$dynamicAnchor"]):
                yield from validator.descend(instance, subschema)
                break
    else:
        with validator.resolver.resolving(dynamicRef) as subschema:
            yield from validator.descend(instance, subschema)


def type(validator, types, instance, schema):
    types = ensure_list(types)

    if not any(validator.is_type(instance, type) for type in types):
        reprs = ", ".join(repr(type) for type in types)
        yield ValidationError(f"{instance!r} is not of type {reprs}")


def properties(validator, properties, instance, schema):
    if not validator.is_type(instance, "object"):
        return

    for property, subschema in properties.items():
        if property in instance:
            yield from validator.descend(
                instance[property],
                subschema,
                path=property,
                schema_path=property,
            )


def required(validator, required, instance, schema):
    if not validator.is_type(instance, "object"):
        return
    for property in required:
        if property not in instance:
            yield ValidationError(f"{property!r} is a required property")


def minProperties(validator, mP, instance, schema):
    if validator.is_type(instance, "object") and len(instance) < mP:
        yield ValidationError(f"{instance!r} does not have enough properties")


def maxProperties(validator, mP, instance, schema):
    if not validator.is_type(instance, "object"):
        return
    if validator.is_type(instance, "object") and len(instance) > mP:
        yield ValidationError(f"{instance!r} has too many properties")


def allOf(validator, allOf, instance, schema):
    for index, subschema in enumerate(allOf):
        yield from validator.descend(instance, subschema, schema_path=index)


def anyOf(validator, anyOf, instance, schema):
    all_errors = []
    for index, subschema in enumerate(anyOf):
        errs = list(validator.descend(instance, subschema, schema_path=index))
        if not errs:
            break
        all_errors.extend(errs)
    else:
        yield ValidationError(
            f"{instance!r} is not valid under any of the given schemas",
            context=all_errors,
        )


def oneOf(validator, oneOf, instance, schema):
    subschemas = enumerate(oneOf)
    all_errors = []
    for index, subschema in subschemas:
        errs = list(validator.descend(instance, subschema, schema_path=index))
        if not errs:
            first_valid = subschema
            break
        all_errors.extend(errs)
    else:
        yield ValidationError(
            f"{instance!r} is not valid under any of the given schemas",
            context=all_errors,
        )

    more_valid = [
        each for _, each in subschemas
        if validator.evolve(schema=each).is_valid(instance)
    ]
    if more_valid:
        more_valid.append(first_valid)
        reprs = ", ".join(repr(schema) for schema in more_valid)
        yield ValidationError(f"{instance!r} is valid under each of {reprs}")


def not_(validator, not_schema, instance, schema):
    if validator.evolve(schema=not_schema).is_valid(instance):
        message = f"{instance!r} should not be valid under {not_schema!r}"
        yield ValidationError(message)


def if_(validator, if_schema, instance, schema):
    if validator.evolve(schema=if_schema).is_valid(instance):
        if "then" in schema:
            then = schema["then"]
            yield from validator.descend(instance, then, schema_path="then")
    elif "else" in schema:
        else_ = schema["else"]
        yield from validator.descend(instance, else_, schema_path="else")


def unevaluatedItems(validator, unevaluatedItems, instance, schema):
    if not validator.is_type(instance, "array"):
        return
    evaluated_item_indexes = find_evaluated_item_indexes_by_schema(
        validator, instance, schema,
    )
    unevaluated_items = [
        item for index, item in enumerate(instance)
        if index not in evaluated_item_indexes
    ]
    if unevaluated_items:
        error = "Unevaluated items are not allowed (%s %s unexpected)"
        yield ValidationError(error % extras_msg(unevaluated_items))


def unevaluatedProperties(validator, unevaluatedProperties, instance, schema):
    if not validator.is_type(instance, "object"):
        return
    evaluated_keys = find_evaluated_property_keys_by_schema(
        validator, instance, schema,
    )
    unevaluated_keys = []
    for property in instance:
        if property not in evaluated_keys:
            for _ in validator.descend(
                instance[property],
                unevaluatedProperties,
                path=property,
                schema_path=property,
            ):
                # FIXME: Include context for each unevaluated property
                #        indicating why it's invalid under the subschema.
                unevaluated_keys.append(property)

    if unevaluated_keys:
        if unevaluatedProperties is False:
            error = "Unevaluated properties are not allowed (%s %s unexpected)"
            yield ValidationError(error % extras_msg(unevaluated_keys))
        else:
            error = (
                "Unevaluated properties are not valid under "
                "the given schema (%s %s unevaluated and invalid)"
            )
            yield ValidationError(error % extras_msg(unevaluated_keys))


def prefixItems(validator, prefixItems, instance, schema):
    if not validator.is_type(instance, "array"):
        return

    for (index, item), subschema in zip(enumerate(instance), prefixItems):
        yield from validator.descend(
            instance=item,
            schema=subschema,
            schema_path=index,
            path=index,
        )



import pytest

from asdf._commands import main


# The test file contains a number of tags that aren't part of the standard.
# This is useful for these tests as we want to handle this case gracefully.
@pytest.mark.filterwarnings("ignore::asdf.exceptions.AsdfConversionWarning")
def test_info_command(capsys, test_data_path):
    file_path = test_data_path / "frames0.asdf"

    assert main.main_from_args(["info", str(file_path)]) == 0
    captured = capsys.readouterr()
    assert "root" in captured.out
    assert "frames" in captured.out
    original_len = len(captured.out.split("\n"))

    assert main.main_from_args(["info", "--max-rows", str(original_len - 5), str(file_path)]) == 0
    captured = capsys.readouterr()
    assert "root" in captured.out
    assert "frames" in captured.out
    new_len = len(captured.out.split("\n"))
    assert new_len < original_len



from io import BytesIO

from asdf._asdf import AsdfFile, open_asdf
from asdf.util import NotSet

__all__ = ["dump", "dumps", "load", "loads"]


def dump(
    tree,
    fp,
    *,
    version=None,
    extensions=None,
    all_array_storage=NotSet,
    all_array_compression=NotSet,
    compression_kwargs=NotSet,
    pad_blocks=False,
    custom_schema=None,
):
    """
    Write a tree to an ASDF file.

    Parameters
    ----------
    tree : object
        The tree to dump.

    fp : str or file-like object
        A file-like object to write the ASDF data to.

    version : str, optional
        Version of the ASDF core schemas to use.  If not specified, the default
        version will be used.

    extensions : object, optional
        Additional extensions to use when reading and writing the file.
        May be an `asdf.extension.Extension` or a `list` of extensions.

    all_array_storage : string, optional
        If provided, override the array storage type of all blocks.

    all_array_compression : string, optional
        If provided, override the array compression type of all blocks.

    compression_kwargs : dict, optional
        If provided, override the compression parameters of all blocks.

    pad_blocks : bool, optional
        If provided, pad all blocks to the nearest multiple of the block size.

    custom_schema : str, optional
        Path to a custom schema file that will be used for a secondary
        validation pass. This can be used to ensure that particular ASDF
        files follow custom conventions beyond those enforced by the
        specification.
    """
    AsdfFile(tree, custom_schema=custom_schema, extensions=extensions).write_to(
        fp,
        version=version,
        all_array_storage=all_array_storage,
        all_array_compression=all_array_compression,
        compression_kwargs=compression_kwargs,
        pad_blocks=pad_blocks,
    )


def dumps(
    tree,
    *,
    version=None,
    extensions=None,
    all_array_storage=NotSet,
    all_array_compression=NotSet,
    compression_kwargs=NotSet,
    pad_blocks=False,
    custom_schema=None,
):
    """
    Write tree to a string.

    Parameters
    ----------
    tree : object
        The tree to dump.

    version : str, optional
        Version of the ASDF core schemas to use.  If not specified, the default
        version will be used.

    extensions : object, optional
        Additional extensions to use when reading and writing the file.
        May be an `asdf.extension.Extension` or a `list` of extensions.

    all_array_storage : string, optional
        If provided, override the array storage type of all blocks.

    all_array_compression : string, optional
        If provided, override the array compression type of all blocks.

    compression_kwargs : dict, optional
        If provided, override the compression parameters of all blocks.

    pad_blocks : bool, optional
        If provided, pad all blocks to the nearest multiple of the block size.

    custom_schema : str, optional
        Path to a custom schema file that will be used for a secondary
        validation pass. This can be used to ensure that particular ASDF
        files follow custom conventions beyond those enforced by the
        specification.

    Returns
    -------
    str
        The ASDF data as a string.
    """
    buff = BytesIO()
    dump(
        tree,
        buff,
        version=version,
        extensions=extensions,
        all_array_storage=all_array_storage,
        all_array_compression=all_array_compression,
        compression_kwargs=compression_kwargs,
        pad_blocks=pad_blocks,
        custom_schema=custom_schema,
    )
    return buff.getvalue()


def load(fp, *, uri=None, validate_checksums=False, extensions=None, custom_schema=None):
    """
    Load the ASDF tree from a file-like object.

    Parameters
    ----------
    fp : str or file-like object
        A file-like object to read the ASDF data from.

    uri : str, optional
        The URI for this ASDF file.  Used to resolve relative
        references against.  If not provided, will be
        automatically determined from the associated file object,
        if possible and if created from `asdf.open`.

    validate_checksums : bool, optional
        If `True`, validate the blocks against their checksums.

    extensions : object, optional
        Additional extensions to use when reading and writing the file.
        May be an `asdf.extension.Extension` or a `list` of extensions.

    custom_schema : str, optional
        Path to a custom schema file that will be used for a secondary
        validation pass. This can be used to ensure that particular ASDF
        files follow custom conventions beyond those enforced by the
        specification.

    Returns
    -------
    object:
        The ASDF tree.
    """
    with open_asdf(
        fp,
        lazy_load=False,
        memmap=False,
        lazy_tree=False,
        uri=uri,
        validate_checksums=validate_checksums,
        extensions=extensions,
        custom_schema=custom_schema,
    ) as af:
        return af.tree


def loads(asdf_string, *, uri=None, validate_checksums=False, extensions=None, custom_schema=None):
    """
    Load the ASDF tree from a string..

    Parameters
    ----------
    asdf_string : str
        A string containing ASDF data.

    uri : str, optional
        The URI for this ASDF file.  Used to resolve relative
        references against.  If not provided, will be
        automatically determined from the associated file object,
        if possible and if created from `asdf.open`.

    validate_checksums : bool, optional
        If `True`, validate the blocks against their checksums.

    extensions : object, optional
        Additional extensions to use when reading and writing the file.
        May be an `asdf.extension.Extension` or a `list` of extensions.

    custom_schema : str, optional
        Path to a custom schema file that will be used for a secondary
        validation pass. This can be used to ensure that particular ASDF
        files follow custom conventions beyond those enforced by the
        specification.

    Returns
    -------
    object:
        The ASDF tree.
    """

    return load(
        BytesIO(asdf_string),
        uri=uri,
        validate_checksums=validate_checksums,
        extensions=extensions,
        custom_schema=custom_schema,
    )



import numpy as np

import asdf


def test_asdf_info_should_not_load_arrays(tmp_path):
    """
    AsdfFile.info should not load array data

    https://github.com/asdf-format/asdf/issues/1553
    """
    fn = tmp_path / "test.asdf"
    tree = dict([(k, np.arange(ord(k))) for k in "abc"])
    asdf.AsdfFile(tree).write_to(fn)

    with asdf.open(fn) as af:
        assert "unloaded" in str(af["b"])
        af.info()
        assert "unloaded" in str(af["b"])



"""
For external blocks, the previous block management
would cache data opened from external files (to return the
same underlying ndarray if the same external block
was referenced more than once). `ExternalBlockCache` is
used here to allow for the same behavior without requiring
the block manager to have a reference to the `AsdfFile`
(that references the block manager).
"""

import os
import urllib

import numpy as np

from asdf import generic_io, util


class UseInternalType:
    pass


UseInternal = UseInternalType()


class ExternalBlockCache:
    def __init__(self):
        self.clear()

    def load(self, base_uri, uri, memmap=False, validate_checksums=False):
        key = util.get_base_uri(uri)
        if key not in self._cache:
            resolved_uri = generic_io.resolve_uri(base_uri, uri)
            # if the uri only has a trailing "#" fragment, strip it
            # this deals with python 3.14 changes where in prior versions
            # urljoin removed this type of fragment
            if resolved_uri.endswith("#"):
                resolved_uri = resolved_uri[:-1]
            if resolved_uri == "" or resolved_uri == base_uri:
                return UseInternal

            from asdf import open as asdf_open

            with asdf_open(
                resolved_uri, "r", lazy_load=False, memmap=False, validate_checksums=validate_checksums
            ) as af:
                blk = af._blocks.blocks[0]
                if memmap and blk.header["compression"] == b"\0\0\0\0":
                    parsed_url = util._patched_urllib_parse.urlparse(resolved_uri)
                    if parsed_url.scheme == "file":
                        # deal with leading slash for windows file://
                        filename = urllib.request.url2pathname(parsed_url.path)
                        arr = np.memmap(filename, np.uint8, "r", blk.data_offset, blk.cached_data.nbytes)
                    else:
                        arr = blk.cached_data
                else:
                    arr = blk.cached_data
            self._cache[key] = arr
        return self._cache[key]

    def clear(self):
        self._cache = {}


def relative_uri_for_index(uri, index):
    # get the os-native separated path for this uri
    path = util._patched_urllib_parse.urlparse(uri).path
    _, filename = os.path.split(path)
    filename = os.path.splitext(filename)[0] + f"{index:04d}.asdf"
    return filename



import re

import numpy as np
import pytest

from asdf import AsdfFile


@pytest.fixture()
def asdf_file():
    tree = {
        "foo": 42,
        "nested": {"foo": 24, "foible": "whoops", "folicle": "yup", "moo": 24},
        "bar": "hello",
        "list": [{"index": 0}, {"index": 1}, {"index": 2}],
    }
    return AsdfFile(tree)


def test_no_arguments(asdf_file):
    result = asdf_file.search()
    assert len(result.paths) == 15
    assert len(result.nodes) == 15


def test_repr(asdf_file):
    result = asdf_file.search()
    assert "foo" in repr(result)
    assert "nested" in repr(result)
    assert "bar" in repr(result)
    assert "list" in repr(result)


def test_single_result(asdf_file):
    result = asdf_file.search("bar")
    assert len(result.paths) == 1
    assert len(result.nodes) == 1
    assert result.node == "hello"
    assert result.path == "root['bar']"
    result.replace("goodbye")
    assert asdf_file["bar"] == "goodbye"


def test_multiple_results(asdf_file):
    result = asdf_file.search("foo")
    assert len(result.paths) == 2
    assert len(result.nodes) == 2
    assert 42 in result.nodes
    assert 24 in result.nodes
    assert "root['foo']" in result.paths
    assert "root['nested']['foo']" in result.paths

    with pytest.raises(RuntimeError, match=r"More than one result"):
        result.path

    with pytest.raises(RuntimeError, match=r"More than one result"):
        result.node

    result.replace(54)
    assert asdf_file["foo"] == 54
    assert asdf_file["nested"]["foo"] == 54


def test_by_key(asdf_file):
    result = asdf_file.search("bar")
    assert result.node == "hello"

    result = asdf_file.search("^b.r$")
    assert result.node == "hello"

    result = asdf_file.search(re.compile("fo[oi]"))
    assert set(result.nodes) == {42, 24, "whoops"}

    result = asdf_file.search(0)
    assert result.node == {"index": 0}


def test_by_type(asdf_file):
    result = asdf_file.search(type_=str)
    assert sorted(result.nodes) == sorted(["hello", "whoops", "yup"])

    result = asdf_file.search(type_="int")
    assert result.nodes == [42, 24, 24, 0, 1, 2]

    result = asdf_file.search(type_="dict|list")
    assert len(result.nodes) == 5

    result = asdf_file.search(type_=re.compile("^i.t$"))
    assert result.nodes == [42, 24, 24, 0, 1, 2]

    with pytest.raises(TypeError, match=r"type must be .*"):
        asdf_file.search(type_=4)


def test_by_value(asdf_file):
    result = asdf_file.search(value=42)
    assert result.node == 42


def test_by_value_with_ndarray():
    """
    Check some edge cases when comparing integers and booleans to numpy arrays.
    """
    tree = {"foo": np.arange(10)}
    af = AsdfFile(tree)
    result = af.search(value=True)
    assert len(result.nodes) == 0
    result = af.search(value=42)
    assert len(result.nodes) == 0


def test_by_filter(asdf_file):
    with pytest.raises(ValueError, match=r"filter must accept 1 or 2 arguments"):
        asdf_file.search(filter_=lambda: True)

    result = asdf_file.search(filter_=lambda n: isinstance(n, int) and n % 2 == 0)
    assert result.nodes == [42, 24, 24, 0, 2]

    result = asdf_file.search(filter_=lambda n, k: k == "foo" and n > 30)
    assert result.node == 42


def test_multiple_conditions(asdf_file):
    result = asdf_file.search("foo", value=24)
    assert len(result.nodes) == 1
    assert result.node == 24
    result.replace(19)
    assert len(result.nodes) == 0
    assert asdf_file["foo"] == 42
    assert asdf_file["nested"]["foo"] == 19


def test_chaining(asdf_file):
    result = asdf_file.search("foo").search(value=24)
    assert len(result.nodes) == 1
    assert result.node == 24
    result.replace(19)
    assert len(result.nodes) == 0
    assert asdf_file["foo"] == 42
    assert asdf_file["nested"]["foo"] == 19


def test_index_operator(asdf_file):
    result = asdf_file.search()["nested"].search("foo")
    assert len(result.nodes) == 1
    assert result.node == 24

    with pytest.raises(TypeError, match=r"This node cannot be indexed"):
        asdf_file.search()["foo"][0]


def test_format(asdf_file):
    result = asdf_file.search()
    original_len = len(repr(result).split("\n"))

    result = result.format(max_rows=original_len - 5)
    new_len = len(repr(result).split("\n"))
    assert new_len < original_len

    result = result.format(max_rows=(None, 5))
    new_len = len(repr(result).split("\n"))
    assert new_len < original_len

    assert repr(result) == repr(result.format())


def test_no_results(asdf_file):
    result = asdf_file.search("missing")
    assert len(result.nodes) == 0
    assert "No results found." in repr(result)
    assert result.node is None
    assert result.path is None
    # Testing no exceptions here:
    result.replace("foo")


def test_recursive_tree():
    tree = {"foo": {"bar": "baz"}}
    af = AsdfFile(tree)
    af.tree["foo"]["nested"] = af.tree["foo"]

    result = af.search()
    assert "(recursive reference)" in repr(result)

    result = af.search("bar")
    assert len(result.nodes) == 1
    assert result.node == "baz"

    result.replace("zap")
    assert af["foo"]["bar"] == "zap"
    assert af["foo"]["nested"]["bar"] == "zap"


def test_search():
    tree = {"foo": 42, "bar": "hello", "baz": np.arange(20)}
    af = AsdfFile(tree)

    result = af.search("foo")
    assert result.node == 42

    result = af.search(type_="ndarray")
    assert (result.node == tree["baz"]).all()

    result = af.search(value="hello")
    assert result.node == "hello"



"""
Implementation of command for displaying differences between two ASDF files.
"""

import argparse
import sys

import jmespath
import numpy as np
from numpy import array_equal

import asdf
from asdf.extension._serialization_context import BlockAccess
from asdf.tagged import Tagged

from .main import Command

__all__ = ["diff"]


NDARRAY_TAG = "core/ndarray"


class Diff(Command):  # pragma: no cover
    """This class is the plugin implementation for the asdftool runner."""

    @classmethod
    def setup_arguments(cls, subparsers):
        epilog = """
examples:
  diff two files:
    asdftool diff file_before.asdf file_after.asdf
  ignore differences in the file's ASDF metadata:
    asdftool diff file_before.asdf file_after.asdf -i '[asdf_library,history]'
  ignore differences in the 'foo' field of all objects in a list:
    asdftool diff file_before.asdf file_after.asdf -i 'path.to.some_list[*].foo'

See https://jmespath.org/ for more information on constructing
JMESPath expressions.
    """.strip()

        parser = subparsers.add_parser(
            "diff",
            description="Report differences between two ASDF files",
            epilog=epilog,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            help="Report differences between two ASDF files",
        )

        parser.add_argument("filenames", metavar="asdf_file", nargs=2, help="The ASDF files to compare.")

        parser.add_argument(
            "-m",
            "--minimal",
            action="store_true",
            help="Show minimal differences between the two files.",
        )

        parser.add_argument(
            "-i",
            "--ignore",
            action="append",
            dest="ignore",
            help="JMESPath expression indicating tree nodes that should be ignored.",
        )

        parser.set_defaults(func=cls.run)
        return parser

    @classmethod
    def run(cls, args):
        return diff(args.filenames, args.minimal, ignore=args.ignore)


class ArrayNode:
    """This class is used to represent unique dummy nodes in the diff tree. In
    general these dummy nodes will be list elements that we want to keep track
    of but not necessarily display. This allows the diff output to be
    cleaner."""

    def __init__(self, name, index):
        self.name = name
        self.index = index

    def __hash__(self):
        return hash(self.name)


class PrintTree:
    """This class is used to remember the nodes in the tree that have already
    been displayed in the diff output.
    """

    def __init__(self):
        self.__tree = {"visited": False, "children": {}}

    def get_print_list(self, node_list):
        at_end = False
        print_list = []
        current = self.__tree
        for node in ["tree", *node_list]:
            if at_end:
                print_list.append(node)
            elif node not in current["children"]:
                print_list.append(node)
                at_end = True
            elif not current["children"][node]["visited"]:
                print_list.append(node)
            else:
                print_list.append(None)
            if not at_end:
                current = current["children"][node]
        return print_list

    def __setitem__(self, node_list, visit):
        if not isinstance(node_list, list):
            msg = "node_list parameter must be an instance of list"
            raise TypeError(msg)
        current = self.__tree
        for node in ["tree", *node_list]:
            if node not in current["children"]:
                current["children"][node] = {"visited": True, "children": {}}
            current = current["children"][node]


class DiffContext:
    """Class that contains context data of the diff to be computed"""

    def __init__(self, asdf0, asdf1, iostream, minimal=False, ignore_ids=None):
        self._seen_id_tuples = set()
        self.asdf0 = asdf0
        self.asdf1 = asdf1
        self.iostream = iostream
        self.minimal = minimal
        self.print_tree = PrintTree()

        if ignore_ids is None:
            self.ignore_ids = set()
        else:
            self.ignore_ids = ignore_ids

        if hasattr(sys.stdout, "isatty") and sys.stdout.isatty():
            RED = "\x1b[31m"
            GREEN = "\x1b[32m"
            RESET = "\x1b[0m"
        else:
            RED = ""
            GREEN = ""
            RESET = ""

        self.RESET_NEWLINE = RESET + "\n"
        self.LIST_MARKER = "-"
        self.THIS_MARKER = GREEN + "> "
        self.THAT_MARKER = RED + "< "

    def should_visit(self, item0, item1):
        key = (id(item0), id(item1))
        if key in self._seen_id_tuples:
            return False
        self._seen_id_tuples.add(key)
        return True


def print_tree_context(diff_ctx, node_list, other, use_marker, last_was_list):
    """Print context information indicating location in ASDF tree."""
    prefix = ""
    marker = diff_ctx.THAT_MARKER if other else diff_ctx.THIS_MARKER
    for node in diff_ctx.print_tree.get_print_list(node_list):
        if node is not None:
            node_ = diff_ctx.LIST_MARKER if isinstance(node, ArrayNode) else str(node) + ":"
            # All of this logic is just to make the display of arrays prettier
            if use_marker:
                line_prefix = " " if last_was_list else marker + prefix[2:]
                line_suffix = "" if node_ == diff_ctx.LIST_MARKER else diff_ctx.RESET_NEWLINE
            else:
                line_prefix = prefix
                line_suffix = diff_ctx.RESET_NEWLINE
            diff_ctx.iostream.write(line_prefix + node_ + line_suffix)
            last_was_list = node_ == diff_ctx.LIST_MARKER
        prefix += "  "
    diff_ctx.print_tree[node_list] = True
    return last_was_list


def print_in_tree(
    diff_ctx, node_list, thing, other, use_marker=False, last_was_list=False, ignore_lwl=False, seen=None
):
    """Recursively print tree context and diff information about object."""
    last_was_list = print_tree_context(diff_ctx, node_list, other, use_marker, last_was_list)

    if seen is None:
        seen = set()
    seen_key = id(thing)
    if seen_key in seen:
        use_marker = not last_was_list or ignore_lwl
        marker = diff_ctx.THAT_MARKER if other else diff_ctx.THIS_MARKER
        prefix = marker + "  " * len(node_list) if use_marker else " "
        diff_ctx.iostream.write(prefix + "(recursive item)" + diff_ctx.RESET_NEWLINE)
        last_was_list = False
        return
    seen.add(seen_key)

    # If tree element is list, recursively print list contents
    if isinstance(thing, list):
        for i, subthing in enumerate(thing):
            key = ArrayNode(f"{node_list[-1]}_{i}", i)
            last_was_list = print_in_tree(
                diff_ctx,
                [*node_list, key],
                subthing,
                other,
                use_marker=True,
                last_was_list=last_was_list,
                ignore_lwl=ignore_lwl,
                seen=seen,
            )
    # If tree element is dictionary, recursively print dictionary contents
    elif isinstance(thing, dict):
        for key in sorted(thing.keys()):
            last_was_list = print_in_tree(
                diff_ctx,
                [*node_list, key],
                thing[key],
                other,
                use_marker=True,
                last_was_list=last_was_list,
                ignore_lwl=ignore_lwl,
                seen=seen,
            )
    # Print difference between leaf objects (no need to recurse further)
    else:
        use_marker = not last_was_list or ignore_lwl
        marker = diff_ctx.THAT_MARKER if other else diff_ctx.THIS_MARKER
        prefix = marker + "  " * len(node_list) if use_marker else " "
        diff_ctx.iostream.write(prefix + str(thing) + diff_ctx.RESET_NEWLINE)
        last_was_list = False
    return last_was_list


def compare_objects(diff_ctx, obj0, obj1, keys=None):
    """Displays diff of two objects if they are not equal"""
    keys = [] if keys is None else keys

    if obj0 != obj1:
        print_in_tree(diff_ctx, keys, obj0, False, ignore_lwl=True)
        print_in_tree(diff_ctx, keys, obj1, True, ignore_lwl=True)


def print_dict_diff(diff_ctx, tree, node_list, keys, other):
    """Recursively traverses dictionary object and displays differences"""
    for key in keys:
        if diff_ctx.minimal:
            nodes = node_list
            key_ = key
        else:
            nodes = [*node_list, key]
            key_ = tree[key]
        use_marker = not diff_ctx.minimal
        print_in_tree(diff_ctx, nodes, key_, other, use_marker=use_marker)


def _load_array(asdf_file, array_dict):
    # the array_dict may not be tagged if the array is inline
    # in this case just use what's in "data"
    if not hasattr(array_dict, "_tag"):
        return array_dict["data"]
    conv = asdf_file.extension_manager.get_converter_for_type(np.ndarray)
    sctx = asdf_file._create_serialization_context(BlockAccess.READ)
    return conv.from_yaml_tree(array_dict, array_dict._tag, sctx)


def _human_list(line, separator="and"):
    """
    Formats a list for human readability.

    Parameters
    ----------
    line : sequence
        A sequence of strings

    separator : string, optional
        The word to use between the last two entries.  Default:
        ``"and"``.

    Returns
    -------
    formatted_list : string

    Examples
    --------
    >>> _human_list(["vanilla", "strawberry", "chocolate"], "or")
    'vanilla, strawberry or chocolate'
    """
    if len(line) == 1:
        return line[0]

    return ", ".join(line[:-1]) + " " + separator + " " + line[-1]


def compare_ndarrays(diff_ctx, array0, array1, keys):
    """Compares two ndarray objects"""
    if isinstance(array0, list):
        array0 = {"data": array0}
    if isinstance(array1, list):
        array1 = {"data": array1}

    ignore_keys = {"source", "data"}
    compare_dicts(diff_ctx, array0, array1, keys, ignore_keys)

    differences = []
    for field in ["shape", "datatype"]:
        if array0.get(field) != array1.get(field):
            differences.append(field)

    value0 = _load_array(diff_ctx.asdf0, array0)
    value1 = _load_array(diff_ctx.asdf1, array1)

    if not array_equal(value0, value1):
        differences.append("contents")

    if differences:
        msg = f"ndarrays differ by {_human_list(differences)}"
        print_in_tree(diff_ctx, keys, msg, False, ignore_lwl=True)
        print_in_tree(diff_ctx, keys, msg, True, ignore_lwl=True)


def both_are_ndarrays(tree0, tree1):
    """Returns True if both inputs correspond to ndarrays, False otherwise"""
    if not (isinstance(tree0, Tagged) and isinstance(tree1, Tagged)):
        return False
    if not (NDARRAY_TAG in tree0._tag and NDARRAY_TAG in tree1._tag):
        return False
    return True


def compare_dicts(diff_ctx, dict0, dict1, keys, ignores=None):
    """Recursively compares two dictionary objects"""
    ignores = set() if ignores is None else ignores

    keys0 = set(dict0.keys()) - ignores
    keys1 = set(dict1.keys()) - ignores
    # Recurse into subtree elements that are shared by both trees
    for key in sorted(keys0 & keys1):
        obj0 = dict0[key]
        obj1 = dict1[key]
        compare_trees(diff_ctx, obj0, obj1, keys=[*keys, key])
    # Display subtree elements existing only in this tree
    print_dict_diff(diff_ctx, dict0, keys, sorted(keys0 - keys1), False)
    # Display subtree elements existing only in that tree
    print_dict_diff(diff_ctx, dict1, keys, sorted(keys1 - keys0), True)


def compare_trees(diff_ctx, tree0, tree1, keys=None):
    """Recursively traverses two ASDF tree and compares them"""
    if not diff_ctx.should_visit(tree0, tree1):
        return

    keys = [] if keys is None else keys

    if id(tree0) in diff_ctx.ignore_ids and id(tree1) in diff_ctx.ignore_ids:
        return

    if both_are_ndarrays(tree0, tree1):
        compare_ndarrays(diff_ctx, tree0, tree1, keys)
    elif isinstance(tree0, dict) and isinstance(tree1, dict):
        compare_dicts(diff_ctx, tree0, tree1, keys)
    elif isinstance(tree0, list) and isinstance(tree1, list):
        for i, (obj0, obj1) in enumerate(zip(tree0, tree1)):
            key = ArrayNode(f"item_{i}", i)
            compare_trees(diff_ctx, obj0, obj1, [*keys, key])
    else:
        compare_objects(diff_ctx, tree0, tree1, keys)


def diff(filenames, minimal, iostream=sys.stdout, ignore=None):
    """
    Compare two ASDF files and write diff output to the stdout
    or the specified I/O stream.

    filenames : list of str
        List of ASDF filenames to compare.  Must be length 2.

    minimal : boolean
        Set to True to forego some pretty-printing to minimize
        the diff output.

    iostream : io.TextIOBase, optional
        Text-mode stream to write the diff, e.g., sys.stdout
        or an io.StringIO instance.  Defaults to stdout.

    ignore : list of str, optional
        List of JMESPath expressions indicating tree nodes that
        should be ignored.
    """
    ignore_expressions = [] if ignore is None else [jmespath.compile(e) for e in ignore]

    try:
        with (
            asdf.open(filenames[0], _force_raw_types=True) as asdf0,
            asdf.open(
                filenames[1],
                _force_raw_types=True,
            ) as asdf1,
        ):
            ignore_ids = set()
            for expression in ignore_expressions:
                for tree in [asdf0.tree, asdf1.tree]:
                    result = expression.search(tree)
                    if result is not None:
                        ignore_ids.add(id(result))
                    if isinstance(result, list):
                        for elem in result:
                            ignore_ids.add(id(elem))
                    elif isinstance(result, dict):
                        for value in result.values():
                            ignore_ids.add(id(value))

            diff_ctx = DiffContext(asdf0, asdf1, iostream, minimal=minimal, ignore_ids=ignore_ids)
            compare_trees(diff_ctx, asdf0.tree, asdf1.tree)

    except ValueError as err:
        raise RuntimeError(str(err)) from err



import numpy as np

import asdf
from asdf import AsdfFile
from asdf._commands import main
from asdf._tests._helpers import assert_tree_match


def test_to_yaml(tmp_path):
    x = np.arange(0, 10, dtype=float)

    tree = {
        "science_data": x,
        "subset": x[3:-3],
        "skipping": x[::2],
        "not_shared": np.arange(10, 0, -1, dtype=np.uint8),
    }

    path = tmp_path / "original.asdf"
    ff = AsdfFile(tree)
    ff.write_to(path)
    with asdf.open(path) as ff2:
        assert len(ff2._blocks.blocks) == 2

    result = main.main_from_args(["to_yaml", str(path)])

    assert result == 0

    files = [p.name for p in tmp_path.iterdir()]

    assert "original.asdf" in files
    assert "original.yaml" in files

    with asdf.open(tmp_path / "original.yaml") as ff:
        assert_tree_match(ff.tree, tree)
        assert len(list(ff._blocks.blocks)) == 0



import collections
import copy
import gc
import weakref

import numpy as np
import pytest

import asdf
from asdf.lazy_nodes import AsdfDictNode, AsdfListNode, AsdfOrderedDictNode, _resolve_af_ref, _to_lazy_node


def test_slice_access():
    af = asdf.AsdfFile()
    node = AsdfListNode([0, 1, 2], weakref.ref(af))
    assert node[0] == 0
    assert node[1] == 1
    assert node[2] == 2
    assert node[:2] == [0, 1]
    assert isinstance(node[:2], AsdfListNode)
    assert node[1:2] == [
        1,
    ]
    assert isinstance(node[1:2], AsdfListNode)
    assert node[:-1] == [0, 1]
    assert isinstance(node[:-1], AsdfListNode)
    assert node[::-1] == [2, 1, 0]
    assert isinstance(node[::-1], AsdfListNode)
    assert node[::2] == [0, 2]
    assert isinstance(node[::2], AsdfListNode)
    assert node[1::2] == [
        1,
    ]
    assert isinstance(node[1::2], AsdfListNode)


def test_nested_node_conversion():
    tree = {
        # lll = list in list in list, etc...
        "lll": [[[0]]],
        "lld": [[{"a": 0}]],
        "ldl": [{"a": [0]}],
        "ldd": [{"a": {"a": [0]}}],
        "dll": {"a": [[0]]},
        "dld": {"a": [{"a": 0}]},
        "ddl": {"a": {"a": [0]}},
        "ddd": {"a": {"a": {"a": 0}}},
    }
    af = asdf.AsdfFile()
    node = AsdfDictNode(tree, weakref.ref(af))
    for key in node:
        obj = node[key]
        for code in key:
            if code == "l":
                assert isinstance(obj, AsdfListNode)
                obj = obj[0]
            else:
                assert isinstance(obj, AsdfDictNode)
                obj = obj["a"]


def test_lazy_tree_ref(tmp_path):
    fn = tmp_path / "test.asdf"

    arr = np.arange(42)
    tree = {
        "a": arr,
        "b": {"c": arr},
        "d": [
            arr,
        ],
    }

    af = asdf.AsdfFile(tree)
    af.write_to(fn)

    with asdf.open(fn, lazy_tree=True) as af:
        assert isinstance(af.tree.data.tagged["a"], asdf.tagged.Tagged)
        assert isinstance(af.tree.data.tagged["b"]["c"], asdf.tagged.Tagged)
        assert isinstance(af.tree.data.tagged["d"][0], asdf.tagged.Tagged)
        assert isinstance(af["b"], AsdfDictNode)
        assert isinstance(af["d"], AsdfListNode)
        np.testing.assert_array_equal(af["a"], arr)
        assert af["a"] is af["b"]["c"]
        assert af["a"] is af["d"][0]


def test_ordered_dict():
    tree = {"a": collections.OrderedDict({"b": [1, 2, collections.OrderedDict({"c": 3})]})}

    af = asdf.AsdfFile()

    node = AsdfDictNode(tree, weakref.ref(af))
    assert isinstance(node["a"], AsdfOrderedDictNode)
    assert isinstance(node["a"]["b"], AsdfListNode)
    assert isinstance(node["a"]["b"][2], AsdfOrderedDictNode)


@pytest.mark.parametrize(
    "NodeClass,data,base",
    [
        (AsdfDictNode, {"a": 1}, collections.abc.Mapping),
        (AsdfListNode, [1, 2], collections.abc.Sequence),
        (AsdfOrderedDictNode, {"a": 1}, collections.OrderedDict),
    ],
)
def test_node_inheritance(NodeClass, data, base):
    node = NodeClass(data)
    assert isinstance(node, base)


@pytest.mark.parametrize(
    "NodeClass,base",
    [
        (AsdfDictNode, dict),
        (AsdfListNode, list),
        (AsdfOrderedDictNode, dict),
    ],
)
def test_node_empty_init(NodeClass, base):
    node = NodeClass()
    assert type(node.tagged) is base


@pytest.mark.parametrize(
    "node,base_type",
    [
        (AsdfDictNode({"a": 1, "b": 2}), dict),
        (AsdfListNode([1, 2, 3]), list),
        (AsdfOrderedDictNode({"a": 1, "b": 2}), collections.OrderedDict),
    ],
)
@pytest.mark.parametrize("copy_operation", [copy.copy, copy.deepcopy])
def test_copy(node, base_type, copy_operation):
    copied_node = copy_operation(node)
    if copy_operation is copy.copy:
        assert type(copied_node) is type(node)
    else:
        assert type(copied_node) is base_type
    assert copied_node == node


@pytest.mark.parametrize(
    "NodeClass,data",
    [
        (AsdfDictNode, {1: "a", 2: "b"}),
        (AsdfListNode, [1, 2]),
        (AsdfOrderedDictNode, collections.OrderedDict({1: "a", 2: "b"})),
    ],
)
def test_node_equality(NodeClass, data):
    node = NodeClass(data)
    assert node == node
    assert not node != node
    assert node == data
    data.pop(1)
    assert node != data


def test_cache_clear_on_close(tmp_path):
    fn = tmp_path / "test.asdf"

    arr = np.arange(42)
    tree = {"a": arr}
    asdf.AsdfFile(tree).write_to(fn)

    with asdf.open(fn, lazy_tree=True) as af:
        # grab a weakref to this array, it should fail
        # to resolve after the with exits
        ref = weakref.ref(af["a"])

    gc.collect()
    assert ref() is None


def test_access_after_del(tmp_path):
    fn = tmp_path / "test.asdf"

    arr = np.arange(42)
    tree = {"a": {"b": arr}}
    asdf.AsdfFile(tree).write_to(fn)

    with asdf.open(fn, lazy_tree=True) as af:
        d = af["a"]

    del af
    gc.collect(2)

    with pytest.raises(asdf.exceptions.AsdfLazyReferenceError, match="Failed to resolve"):
        d["b"]


def test_lazy_tree_option(tmp_path):
    fn = tmp_path / "test.asdf"

    arr = np.arange(42)
    tree = {"a": {"b": arr}}
    asdf.AsdfFile(tree).write_to(fn)

    with asdf.open(fn, lazy_tree=True) as af:
        assert isinstance(af["a"], AsdfDictNode)

    with asdf.open(fn, lazy_tree=False) as af:
        assert not isinstance(af["a"], AsdfDictNode)

    # test default (False)
    with asdf.open(fn) as af:
        assert not isinstance(af["a"], AsdfDictNode)

    with asdf.config_context() as cfg:
        cfg.lazy_tree = True
        with asdf.open(fn) as af:
            assert isinstance(af["a"], AsdfDictNode)
        cfg.lazy_tree = False
        with asdf.open(fn) as af:
            assert not isinstance(af["a"], AsdfDictNode)


def test_resolve_af_ref():
    with pytest.raises(asdf.exceptions.AsdfLazyReferenceError, match="Failed to resolve"):
        _resolve_af_ref(None)

    af = asdf.AsdfFile()
    af_ref = weakref.ref(af)
    assert _resolve_af_ref(af_ref) is af

    del af
    gc.collect(2)

    with pytest.raises(asdf.exceptions.AsdfLazyReferenceError, match="Failed to resolve"):
        _resolve_af_ref(af_ref)


@pytest.mark.parametrize(
    "NodeClass,data",
    [
        (AsdfDictNode, {1: "a", 2: "b"}),
        (AsdfListNode, [1, 2]),
        (AsdfOrderedDictNode, collections.OrderedDict({1: "a", 2: "b"})),
        (int, 1),  # a non-wrappable class
    ],
)
def test_to_lazy_node(NodeClass, data):
    node = _to_lazy_node(data, None)
    assert isinstance(node, NodeClass)


def test_lazy_node_treeutil_support():
    af = asdf.AsdfFile()
    af_ref = weakref.ref(af)
    tree = {
        "ordered_dict": AsdfOrderedDictNode({"a": 1}, af_ref),
        "dict": AsdfDictNode({"b": 2}, af_ref),
        "list": AsdfListNode([3, 4], af_ref),
    }
    seen_ints = set()

    def callback(node):
        if isinstance(node, int):
            seen_ints.add(node)

    asdf.treeutil.walk_and_modify(tree, callback)

    assert seen_ints == set([1, 2, 3, 4])


@pytest.fixture()
def cache_test_tree_path(tmp_path):
    my_array = np.arange(3, dtype="uint8")
    my_list = [my_array, my_array]
    tree = {"a": my_list, "b": my_list}
    af = asdf.AsdfFile(tree)
    fn = tmp_path / "test.asdf"
    af.write_to(fn)
    return fn


def test_cache_resolves_ref(cache_test_tree_path):
    with asdf.open(cache_test_tree_path, lazy_tree=True) as af:
        # since 'a' and 'b' were the same list when the file was saved
        # they should be the same list on read
        assert af["a"] is af["b"]
        # same for the arrays in the list
        assert af["a"][0] is af["a"][1]


def test_cache_frees_deleted_object(cache_test_tree_path):
    with asdf.open(cache_test_tree_path, lazy_tree=True) as af:
        # load 1 of the 2 lists
        l0 = af["a"]
        # grab a weakref to the list (to not hold onto the list)
        lref = weakref.ref(l0)
        # now delete all references to the list (including the one in the tree)
        del l0, af.tree["a"]
        # trigger garbage collection
        gc.collect(2)
        # check that the weakref fails to resolve (so the list was freed)
        assert lref() is None
        # and we can no longer access 'a'
        with pytest.raises(KeyError, match="'a'"):
            af["a"]
        # but can get 'b'
        assert af["b"][0] is af["b"][1]


def test_cache_non_weakref():
    """
    Test that an object that cannot weak reference can be cached
    """
    tagged_node = {}
    obj = complex(1, 1)
    cache_item = asdf.lazy_nodes._TaggedObjectCacheItem(tagged_node, obj)
    del obj
    gc.collect(2)
    assert cache_item.custom_object == complex(1, 1)


@pytest.fixture(params=[True, False, None], ids=["lazy", "not-lazy", "undefined"])
def lazy_test_class(request):
    class Foo:
        def __init__(self, data):
            self.data = data

    tag_uri = "asdf://somewhere.org/tags/foo-1.0.0"

    class FooConverter:
        tags = [tag_uri]
        types = [Foo]

        def to_yaml_tree(self, obj, tag, ctx):
            return obj.data

        def from_yaml_tree(self, node, tag, ctx):
            return Foo(node)

    lazy = request.param
    if lazy is not None:
        FooConverter.lazy = lazy
    # also set lazy on the class to pass it to the test
    Foo.lazy = lazy

    class FooExtension:
        extension_uri = "asdf://somewhere.org/extensions/minimum-1.0.0"
        converters = [FooConverter()]
        tags = [tag_uri]

    with asdf.config_context() as cfg:
        cfg.add_extension(FooExtension())
        yield Foo


def test_lazy_converter(tmp_path, lazy_test_class):
    obj = lazy_test_class({"a": 1})

    fn = tmp_path / "test.asdf"

    af = asdf.AsdfFile({"obj": obj})
    af.write_to(fn)

    with asdf.open(fn, lazy_tree=True) as af:
        if lazy_test_class.lazy is None or not lazy_test_class.lazy:
            target_class = dict
        else:
            target_class = AsdfDictNode
        assert isinstance(af["obj"].data, target_class)


@pytest.fixture()
def lazy_generator_class(request):

    class Foo:
        def __init__(self, data=None):
            self.data = data or {}

    tag_uri = "asdf://somewhere.org/tags/foo-1.0.0"

    class FooConverter:
        tags = [tag_uri]
        types = [Foo]
        lazy = True

        def to_yaml_tree(self, obj, tag, ctx):
            return obj.data

        def from_yaml_tree(self, node, tag, ctx):
            obj = Foo()
            yield obj
            obj.data = node

    class FooExtension:
        extension_uri = "asdf://somewhere.org/extensions/minimum-1.0.0"
        converters = [FooConverter()]
        tags = [tag_uri]

    with asdf.config_context() as cfg:
        cfg.add_extension(FooExtension())
        yield Foo


def test_lazy_generator_converter(tmp_path, lazy_generator_class):
    """
    Test that a converter that returns a generator is not lazy
    (even if it's marked as lazy).
    """
    obj = lazy_generator_class({"a": 1})

    fn = tmp_path / "test.asdf"

    af = asdf.AsdfFile({"obj": obj})
    af.write_to(fn)

    with asdf.open(fn, lazy_tree=True) as af:
        assert isinstance(af["obj"].data, dict)


def test_lazy_copy(tmp_path):
    """
    Test that copying an AsdfFile instance with a lazy
    tree doesn't result in the copy retaining references
    to the instance.
    """
    fn = tmp_path / "test.asdf"
    obj = asdf.tags.core.IntegerType(1)
    tree = {"a": {"b": obj}}
    # make a recursive structure
    tree["a"]["c"] = tree["a"]

    asdf.AsdfFile(tree).write_to(fn)

    with asdf.open(fn, lazy_tree=True) as af:
        af2 = af.copy()

    del af
    gc.collect(2)
    assert af2["a"]["b"] == obj
    assert af2["a"]["c"]["b"] is af2["a"]["b"]



import os

import pytest

from asdf import config_context
from asdf._asdf import AsdfFile, open_asdf
from asdf._entry_points import get_extensions
from asdf._tests._helpers import assert_tree_match
from asdf.exceptions import AsdfWarning
from asdf.extension import ExtensionProxy
from asdf.testing.helpers import yaml_to_asdf
from asdf.versioning import AsdfVersion


def test_no_warnings_get_extensions():
    """
    Smoke test for changes to the `importlib.metadata` entry points API.
    """

    get_extensions()


class TestExtension:
    __test__ = False

    def __init__(
        self,
        extension_uri=None,
        legacy_class_names=None,
        asdf_standard_requirement=None,
        types=None,
        tag_mapping=None,
        url_mapping=None,
    ):
        self._extension_uri = extension_uri
        self._legacy_class_names = set() if legacy_class_names is None else legacy_class_names
        self._asdf_standard_requirement = asdf_standard_requirement
        self._types = [] if types is None else types
        self._tag_mapping = [] if tag_mapping is None else tag_mapping
        self._url_mapping = [] if url_mapping is None else url_mapping

    @property
    def types(self):
        return self._types

    @property
    def tag_mapping(self):
        return self._tag_mapping

    @property
    def url_mapping(self):
        return self._url_mapping

    @property
    def extension_uri(self):
        return self._extension_uri

    @property
    def legacy_class_names(self):
        return self._legacy_class_names

    @property
    def asdf_standard_requirement(self):
        return self._asdf_standard_requirement


def test_asdf_file_version():
    with config_context() as config:
        config.default_version = "1.2.0"

        af = AsdfFile()
        assert af.version == AsdfVersion("1.2.0")
        assert af.version_string == "1.2.0"

        af = AsdfFile(version="1.3.0")
        assert af.version == AsdfVersion("1.3.0")
        assert af.version_string == "1.3.0"

        af = AsdfFile(version=AsdfVersion("1.3.0"))
        assert af.version == AsdfVersion("1.3.0")
        assert af.version_string == "1.3.0"

        with pytest.raises(ValueError, match=r"ASDF Standard version .* is not supported by asdf==.*"):
            AsdfFile(version="0.5.4")

        with pytest.raises(ValueError, match=r"ASDF Standard version .* is not supported by asdf==.*"):
            AsdfFile(version=AsdfVersion("0.5.4"))

        af = AsdfFile()

        af.version = "1.3.0"
        assert af.version == AsdfVersion("1.3.0")
        assert af.version_string == "1.3.0"

        af.version = AsdfVersion("1.4.0")
        assert af.version == AsdfVersion("1.4.0")
        assert af.version_string == "1.4.0"

        with pytest.raises(ValueError, match=r"ASDF Standard version .* is not supported by asdf==.*"):
            af.version = "0.5.4"

        with pytest.raises(ValueError, match=r"ASDF Standard version .* is not supported by asdf==.*"):
            af.version = AsdfVersion("2.5.4")


def test_asdf_file_extensions():
    af = AsdfFile()
    assert af.extensions == []

    extension = TestExtension(extension_uri="asdf://somewhere.org/extensions/foo-1.0")

    for arg in ([extension], extension):
        af = AsdfFile(extensions=arg)
        assert af.extensions == [ExtensionProxy(extension)]

        af = AsdfFile()
        af.extensions = arg
        assert af.extensions == [ExtensionProxy(extension)]

    msg = r"[The extensions parameter must be an extension.*, Extension must implement the Extension interface]"
    for arg in (object(), [object()]):
        with pytest.raises(TypeError, match=msg):
            AsdfFile(extensions=arg)


def test_asdf_file_version_requirement():
    extension_with_requirement = TestExtension(
        extension_uri="asdf://somewhere.org/extensions/foo-1.0",
        asdf_standard_requirement="==1.5.0",
    )

    # No warnings if the requirement is fulfilled:
    AsdfFile(version="1.5.0", extensions=[extension_with_requirement])

    # Version doesn't match the requirement, so we should see a warning
    # and the extension should not be enabled:
    with pytest.warns(AsdfWarning, match=r"does not support ASDF Standard 1.4.0"):
        af = AsdfFile(version="1.4.0", extensions=[extension_with_requirement])
        assert ExtensionProxy(extension_with_requirement) not in af.extensions

    # Version initially matches the requirement, but changing
    # the version on the AsdfFile invalidates it:
    af = AsdfFile(version="1.5.0", extensions=[extension_with_requirement])
    assert ExtensionProxy(extension_with_requirement) in af.extensions
    with pytest.warns(AsdfWarning, match=r"does not support ASDF Standard 1.4.0"):
        af.version = "1.4.0"
    assert ExtensionProxy(extension_with_requirement) not in af.extensions

    # Extension registered with the config should not provoke
    # a warning:
    with config_context() as config:
        config.add_extension(extension_with_requirement)
        af = AsdfFile(version="1.4.0")

        # ... unless the user explicitly requested the invalid extension:
        with pytest.warns(AsdfWarning, match=r"does not support ASDF Standard 1.4.0"):
            af = AsdfFile(version="1.4.0", extensions=[extension_with_requirement])


def test_open_asdf_extensions(tmp_path):
    extension = TestExtension(extension_uri="asdf://somewhere.org/extensions/foo-1.0")

    path = str(tmp_path / "test.asdf")

    with AsdfFile() as af:
        af.write_to(path)

    with open_asdf(path) as af:
        assert af.extensions == []

    for arg in ([extension], extension):
        with open_asdf(path, extensions=arg) as af:
            assert af.extensions == [ExtensionProxy(extension)]

    msg = r"[The extensions parameter must be an extension.*, Extension must implement the Extension interface]"
    for arg in (object(), [object()]):
        with pytest.raises(TypeError, match=msg), open_asdf(path, extensions=arg) as af:
            pass


def test_reading_extension_metadata():
    extension_with_uri = ExtensionProxy(
        TestExtension(extension_uri="asdf://somewhere.org/extensions/foo-1.0"),
        package_name="foo",
        package_version="1.2.3",
    )
    extension_without_uri = ExtensionProxy(
        TestExtension(),
        package_name="foo",
        package_version="1.2.3",
    )
    extension_with_legacy_class_names = ExtensionProxy(
        TestExtension(
            extension_uri="asdf://somewhere.org/extensions/with-legacy-1.0",
            legacy_class_names={"some.legacy.class.Name"},
        ),
        package_name="foo",
        package_version="1.2.3",
    )

    with config_context() as config:
        config.add_extension(extension_with_uri)
        config.add_extension(extension_without_uri)
        config.add_extension(extension_with_legacy_class_names)

        # Test missing history:
        content = "foo: bar"
        buff = yaml_to_asdf(content)
        open_asdf(buff)

        # Test the old history format:
        content = """
history:
  - !core/history_entry-1.0.0
    description: Once upon a time, there was a carnivorous panda.
  - !core/history_entry-1.0.0
    description: This entry intentionally left blank.
foo: bar
        """
        buff = yaml_to_asdf(content, version="1.0.0")
        open_asdf(buff)

        # Test matching by URI:
        content = """
history:
  extensions:
    - !core/extension_metadata-1.0.0
      extension_uri: asdf://somewhere.org/extensions/foo-1.0
      extension_class: some.unrecognized.extension.class.Name
        """
        buff = yaml_to_asdf(content)
        open_asdf(buff)

        # Test matching by legacy class name:
        content = """
history:
  extensions:
    - !core/extension_metadata-1.0.0
      extension_class: some.legacy.class.Name
        """
        buff = yaml_to_asdf(content)
        open_asdf(buff)

        # Warn when the URI is missing, even if there's
        # a class name match:
        content = f"""
history:
  extensions:
    - !core/extension_metadata-1.0.0
      extension_uri: some-missing-URI
      extension_class: {extension_with_uri.class_name}
        """
        buff = yaml_to_asdf(content)
        with pytest.warns(AsdfWarning, match=r"URI 'some-missing-URI'"):
            open_asdf(buff)

        # Warn when the class name is missing:
        content = """
history:
  extensions:
    - !core/extension_metadata-1.0.0
      extension_class: some.missing.class.Name
        """
        buff = yaml_to_asdf(content)
        with pytest.warns(AsdfWarning, match=r"class 'some.missing.class.Name'"):
            open_asdf(buff)

        # Warn when the package version is older:
        content = """
history:
  extensions:
    - !core/extension_metadata-1.0.0
      extension_uri: asdf://somewhere.org/extensions/foo-1.0
      extension_class: some.class.Name
      software: !core/software-1.0.0
        name: foo
        version: 9.2.4
        """
        buff = yaml_to_asdf(content)
        with pytest.warns(AsdfWarning, match=r"older package"):
            open_asdf(buff)

        # Shouldn't warn when the package version is later:
        content = """
history:
  extensions:
    - !core/extension_metadata-1.0.0
      extension_uri: asdf://somewhere.org/extensions/foo-1.0
      extension_class: some.class.Name
      software: !core/software-1.0.0
        name: foo
        version: 0.1.2
        """
        buff = yaml_to_asdf(content)
        open_asdf(buff)

        # Shouldn't receive a warning when the package
        # name changes, even if the version is later:
        content = """
history:
  extensions:
    - !core/extension_metadata-1.0.0
      extension_uri: asdf://somewhere.org/extensions/foo-1.0
      extension_class: some.class.Name
      software: !core/software-1.0.0
        name: bar
        version: 9.4.5
        """
        buff = yaml_to_asdf(content)
        open_asdf(buff)


def test_bad_input(tmp_path):
    """Make sure these functions behave properly with bad input"""
    text_file = str(tmp_path / "test.txt")

    with open(text_file, "w") as fh:
        fh.write("I <3 ASDF!!!!!")

    with pytest.raises(
        ValueError,
        match=r"Does not appear to be a ASDF file.",
    ):
        open_asdf(text_file)


def test_unclosed_file(tmp_path):
    """
    Issue #1006 reported an unclosed file when asdf.open fails
    This is a regression test for the fix in PR #1221
    """
    path = tmp_path / "empty.asdf"
    path.touch()

    with (
        pytest.raises(
            ValueError,
            match=r"Does not appear to be a ASDF file.",
        ),
        open_asdf(path),
    ):
        pass


def test_fsspec(tmp_path):
    """
    Issue #1146 reported errors when opening a fsspec 'file'
    This is a regression test for the fix in PR #1226
    """
    fsspec = pytest.importorskip("fsspec")

    tree = {"a": 1}
    af = AsdfFile(tree)
    fn = tmp_path / "test.asdf"
    af.write_to(fn)

    with fsspec.open(fn) as f:
        af = open_asdf(f)
        assert_tree_match(tree, af.tree)


@pytest.mark.remote_data()
def test_fsspec_http(httpserver):
    """
    Issue #1146 reported errors when opening a fsspec url (using the http
    filesystem)
    This is a regression test for the fix in PR #1228
    """
    fsspec = pytest.importorskip("fsspec")

    tree = {"a": 1}
    af = AsdfFile(tree)
    path = os.path.join(httpserver.tmpdir, "test")
    af.write_to(path)

    fn = httpserver.url + "test"
    with fsspec.open(fn) as f:
        af = open_asdf(f)
        assert_tree_match(tree, af.tree)



import numpy as np
import pytest

import asdf
from asdf import get_config
from asdf.extension import ExtensionManager
from asdf.extension._serialization_context import BlockAccess, SerializationContext


def test_serialization_context():
    extension_manager = ExtensionManager([])
    context = SerializationContext("1.4.0", extension_manager, "file://test.asdf", None)
    assert context.version == "1.4.0"
    assert context.extension_manager is extension_manager
    assert context._extensions_used == set()

    extension = get_config().extensions[0]
    context._mark_extension_used(extension)
    assert context._extensions_used == {extension}
    context._mark_extension_used(extension)
    assert context._extensions_used == {extension}
    context._mark_extension_used(extension.delegate)
    assert context._extensions_used == {extension}

    assert context.url == context._url == "file://test.asdf"

    with pytest.raises(TypeError, match=r"Extension must implement the Extension interface"):
        context._mark_extension_used(object())

    with pytest.raises(ValueError, match=r"ASDF Standard version .* is not supported by asdf==.*"):
        SerializationContext("0.5.4", extension_manager, None, None)


def test_get_block_data_callback(tmp_path):
    fn = tmp_path / "test.asdf"

    # make a file with 2 blocks
    arr0 = np.arange(3, dtype="uint8")
    arr1 = np.arange(10, dtype="uint8")
    asdf.AsdfFile({"arr0": arr0, "arr1": arr1}).write_to(fn)

    with asdf.open(fn) as af:
        context = af._create_serialization_context()
        with pytest.raises(NotImplementedError, match="abstract"):
            context.get_block_data_callback(0)

        op_ctx = af._create_serialization_context(BlockAccess.READ)
        cb0 = op_ctx.get_block_data_callback(0)

        # getting the same callback should pass and return the same object
        assert op_ctx.get_block_data_callback(0) is cb0

        # since we accessed block 0 we shouldn't be allowed to access block 1
        with pytest.raises(OSError, match=r"Converters accessing >1.*"):
            op_ctx.get_block_data_callback(1)

        # unless we use a key
        key = op_ctx.generate_block_key()
        cb1 = op_ctx.get_block_data_callback(1, key)
        assert op_ctx.get_block_data_callback(1, key) is cb1

        # we don't know the order of blocks, so find which block
        # was used for which array by looking at the size
        d0 = cb0()
        d1 = cb1()
        if d0.size == arr1.size:
            arr0, arr1 = arr1, arr0
        np.testing.assert_array_equal(d0, arr0)
        np.testing.assert_array_equal(d1, arr1)

        for access in (BlockAccess.NONE, BlockAccess.WRITE):
            op_ctx = af._create_serialization_context(access)
            with pytest.raises(NotImplementedError, match="abstract"):
                op_ctx.get_block_data_callback(0)


def test_find_available_block_index():
    af = asdf.AsdfFile()
    context = af._create_serialization_context()

    def cb():
        return np.arange(3, dtype="uint8")

    with pytest.raises(NotImplementedError, match="abstract"):
        context.find_available_block_index(cb)

    class Foo:
        pass

    op_ctx = af._create_serialization_context(BlockAccess.WRITE)
    op_ctx.assign_object(Foo())
    assert op_ctx.find_available_block_index(cb) == 0

    for access in (BlockAccess.NONE, BlockAccess.READ):
        op_ctx = af._create_serialization_context(access)
        with pytest.raises(NotImplementedError, match="abstract"):
            op_ctx.find_available_block_index(cb)


def test_generate_block_key():
    af = asdf.AsdfFile()
    context = af._create_serialization_context()

    with pytest.raises(NotImplementedError, match="abstract"):
        context.generate_block_key()

    class Foo:
        pass

    obj = Foo()
    op_ctx = af._create_serialization_context(BlockAccess.WRITE)
    op_ctx.assign_object(obj)
    key = op_ctx.generate_block_key()
    assert key._is_valid()
    assert key._matches_object(obj)

    obj = Foo()
    op_ctx = af._create_serialization_context(BlockAccess.READ)
    # because this test generates but does not assign a key
    # it should raise an exception
    with pytest.raises(OSError, match=r"Converter generated a key.*"):
        key = op_ctx.generate_block_key()
        # the key does not yet have an assigned object
        assert not key._is_valid()
        op_ctx.assign_blocks()


@pytest.mark.parametrize("block_access", [None, *list(BlockAccess)])
def test_get_set_array_storage(block_access):
    af = asdf.AsdfFile()
    if block_access is None:
        context = af._create_serialization_context()
    else:
        context = af._create_serialization_context(block_access)
    arr = np.zeros(3)
    storage = "external"
    assert af.get_array_storage(arr) != storage
    context.set_array_storage(arr, storage)
    assert af.get_array_storage(arr) == storage
    assert context.get_array_storage(arr) == storage


@pytest.mark.parametrize("block_access", [None, *list(BlockAccess)])
def test_get_set_array_compression(block_access):
    af = asdf.AsdfFile()
    if block_access is None:
        context = af._create_serialization_context()
    else:
        context = af._create_serialization_context(block_access)
    arr = np.zeros(3)
    compression = "bzp2"
    kwargs = {"a": 1}
    assert af.get_array_compression(arr) != compression
    assert af.get_array_compression_kwargs(arr) != kwargs
    context.set_array_compression(arr, compression, **kwargs)
    assert af.get_array_compression(arr) == compression
    assert af.get_array_compression_kwargs(arr) == kwargs
    assert context.get_array_compression(arr) == compression
    assert context.get_array_compression_kwargs(arr) == kwargs


def test_get_set_array_save_base():
    af = asdf.AsdfFile()
    context = af._create_serialization_context()
    arr = np.zeros(3)
    cfg = asdf.get_config()
    save_base = cfg.default_array_save_base
    assert af.get_array_save_base(arr) == save_base
    assert context.get_array_save_base(arr) == save_base

    save_base = not save_base
    context.set_array_save_base(arr, save_base)
    assert af.get_array_save_base(arr) == save_base
    assert context.get_array_save_base(arr) == save_base

    save_base = not save_base
    af.set_array_save_base(arr, save_base)
    assert af.get_array_save_base(arr) == save_base
    assert context.get_array_save_base(arr) == save_base

    af.set_array_save_base(arr, None)
    assert af.get_array_save_base(arr) is None
    assert context.get_array_save_base(arr) is None


@pytest.mark.parametrize("value", [1, "true"])
def test_invalid_set_array_save_base(value):
    af = asdf.AsdfFile()
    context = af._create_serialization_context()
    arr = np.zeros(3)
    with pytest.raises(ValueError, match="save_base must be a bool or None"):
        af.set_array_save_base(arr, value)
    with pytest.raises(ValueError, match="save_base must be a bool or None"):
        context.set_array_save_base(arr, value)



import sys

import pytest

from asdf import _entry_points
from asdf._version import version as asdf_package_version
from asdf.exceptions import AsdfWarning
from asdf.extension import ExtensionProxy
from asdf.resource import ResourceMappingProxy

# The standard library importlib.metadata returns duplicate entrypoints
# for all python versions up to and including 3.11
# https://github.com/python/importlib_metadata/issues/410#issuecomment-1304258228
# see PR https://github.com/asdf-format/asdf/pull/1260
# see issue https://github.com/asdf-format/asdf/issues/1254
if sys.version_info >= (3, 12):
    import importlib.metadata as metadata
else:
    import importlib_metadata as metadata


@pytest.fixture()
def mock_entry_points():
    return []


@pytest.fixture(autouse=True)
def _monkeypatch_entry_points(monkeypatch, mock_entry_points):
    def patched_entry_points(*, group):
        for candidate_group, name, func_name in mock_entry_points:
            if candidate_group == group:
                point = metadata.EntryPoint(name=name, group=group, value=func_name)
                vars(point).update(dist=metadata.distribution("asdf"))

                yield point

    monkeypatch.setattr(_entry_points, "entry_points", patched_entry_points)


def resource_mappings_entry_point_successful():
    return [
        {"http://somewhere.org/schemas/foo-1.0.0": b"foo"},
        {"http://somewhere.org/schemas/bar-1.0.0": b"bar"},
    ]


def resource_mappings_entry_point_failing():
    msg = "NOPE"
    raise Exception(msg)


def resource_mappings_entry_point_bad_element():
    return [
        {"http://somewhere.org/schemas/baz-1.0.0": b"baz"},
        object(),
        {"http://somewhere.org/schemas/foz-1.0.0": b"foz"},
    ]


def test_get_resource_mappings(mock_entry_points):
    mock_entry_points.append(
        (
            "asdf.resource_mappings",
            "successful",
            "asdf._tests.test_entry_points:resource_mappings_entry_point_successful",
        ),
    )
    mappings = _entry_points.get_resource_mappings()
    assert len(mappings) == 2
    for m in mappings:
        assert isinstance(m, ResourceMappingProxy)
        assert m.package_name == "asdf"
        assert m.package_version == asdf_package_version

    mock_entry_points.clear()
    mock_entry_points.append(
        ("asdf.resource_mappings", "failing", "asdf._tests.test_entry_points:resource_mappings_entry_point_failing"),
    )
    with pytest.warns(AsdfWarning, match=r"Exception: NOPE"):
        mappings = _entry_points.get_resource_mappings()
    assert len(mappings) == 0

    mock_entry_points.clear()
    mock_entry_points.append(
        (
            "asdf.resource_mappings",
            "bad_element",
            "asdf._tests.test_entry_points:resource_mappings_entry_point_bad_element",
        ),
    )
    with pytest.warns(AsdfWarning, match=r"TypeError: Resource mapping must implement the Mapping interface"):
        mappings = _entry_points.get_resource_mappings()
    assert len(mappings) == 2


class MinimumExtension:
    def __init__(self, extension_uri):
        self._extension_uri = extension_uri

    @property
    def extension_uri(self):
        return self._extension_uri


def extensions_entry_point_successful():
    return [
        MinimumExtension("http://somewhere.org/extensions/foo-1.0"),
        MinimumExtension("http://somewhere.org/extensions/bar-1.0"),
    ]


def extensions_entry_point_failing():
    msg = "NOPE"
    raise Exception(msg)


def extensions_entry_point_bad_element():
    return [
        MinimumExtension("http://somewhere.org/extensions/baz-1.0"),
        object(),
        MinimumExtension("http://somewhere.org/extensions/foz-1.0"),
    ]


class LegacyExtension:
    types = []
    tag_mapping = []
    url_mapping = []


class FauxLegacyExtension:
    pass


def test_get_extensions(mock_entry_points):
    mock_entry_points.append(
        ("asdf.extensions", "successful", "asdf._tests.test_entry_points:extensions_entry_point_successful"),
    )
    extensions = _entry_points.get_extensions()
    assert len(extensions) == 2
    for e in extensions:
        assert isinstance(e, ExtensionProxy)
        assert e.package_name == "asdf"
        assert e.package_version == asdf_package_version

    mock_entry_points.clear()
    mock_entry_points.append(
        ("asdf.extensions", "failing", "asdf._tests.test_entry_points:extensions_entry_point_failing"),
    )
    with pytest.warns(AsdfWarning, match=r"Exception: NOPE"):
        extensions = _entry_points.get_extensions()
    assert len(extensions) == 0

    mock_entry_points.clear()
    mock_entry_points.append(
        ("asdf.extensions", "bad_element", "asdf._tests.test_entry_points:extensions_entry_point_bad_element"),
    )
    with pytest.warns(
        AsdfWarning,
        match=r"TypeError: Extension must implement the Extension interface",
    ):
        extensions = _entry_points.get_extensions()
    assert len(extensions) == 2

    mock_entry_points.clear()
    mock_entry_points.append(("asdf_extensions", "legacy", "asdf._tests.test_entry_points:LegacyExtension"))
    extensions = _entry_points.get_extensions()
    assert len(extensions) == 0  # asdf_extensions is no longer supported

    mock_entry_points.clear()
    mock_entry_points.append(("asdf.extensions", "failing", "asdf._tests.test_entry_points:FauxLegacyExtension"))
    with pytest.warns(AsdfWarning, match=r"TypeError"):
        extensions = _entry_points.get_extensions()
    assert len(extensions) == 0



import json
import os
import subprocess
import urllib.request
from contextlib import contextmanager
from itertools import groupby
from pathlib import Path

import pytest
import pytest_remotedata
from common import assert_file_correct, generate_file
from packaging.version import Version

import asdf
from asdf.versioning import AsdfVersion

# Strange version present on pypi that doesn't parse as a Version
BAD_VERSIONS = {"0"}

# Minimum library version to read files produced by the current
# version of the code.  We're not maintaining < 2.7.x and bugs in older
# versions prevent valid files from being read.
MIN_VERSION_NEW_FILES = Version("2.7.0")

# Minimum library version to produce files read by the current
# version of the code.  Earlier versions aren't able to generate
# files for all the ASDF Standard versions that they claim to support.
MIN_VERSION_OLD_FILES = Version("2.5.0")

# Minimum library version that is capable of correctly writing
# files that follow ASDF Standard 1.0.0.
# ASDF library versions older than this version would sometimes write newer
# tags that were not listed in the supported standard
MIN_VERSION_1_0_0_STANDARD = Version("2.7.0")

GENERATE_SCRIPT_PATH = Path(__file__).parent / "generate_file.py"
ASSERT_SCRIPT_PATH = Path(__file__).parent / "assert_file_correct.py"

# Starting at this version ndarray tag version 1.1.0 is supported
# this is required for asdf standard 1.6.0 or later
NDARRAY_1_1_0_ASDF_VERSION = Version("2.14.0")
NDARRAY_1_1_0_STANDARD_VERSION = AsdfVersion("1.6.0")


@contextmanager
def internet_temporarily_enabled(verbose=False):
    """
    Context manager that temporarily enables pytest_remotedata
    internet.
    """
    initially_disabled = pytest_remotedata.disable_internet.INTERNET_OFF

    pytest_remotedata.disable_internet.turn_on_internet(verbose=verbose)
    try:
        yield
    finally:
        if initially_disabled:
            pytest_remotedata.disable_internet.turn_off_internet(verbose=verbose)


def fetch_package_versions(package_name):
    """
    Request a package's available versions from pypi.org metadata.
    """
    content = urllib.request.urlopen(f"https://pypi.org/pypi/{package_name}/json").read()
    version_strings = json.loads(content)["releases"].keys()
    return [
        Version(v)
        for v in version_strings
        if v not in BAD_VERSIONS and (Version(v) >= MIN_VERSION_NEW_FILES or Version(v) >= MIN_VERSION_OLD_FILES)
    ]


def fetch_latest_patch_versions(package_name):
    """
    Return the latest patch version within each of the package's
    minor versions.
    """

    def key_fn(v):
        return v.release[0:2]

    versions = sorted(fetch_package_versions(package_name), key=key_fn)
    return [max(group) for _, group in groupby(versions, key=key_fn)]


# Enable internet here, otherwise pytest_remotedata will complain
# (and @pytest.mark.remote_data doesn't work on non-test methods).
with internet_temporarily_enabled():
    PATCH_VERSIONS = sorted(fetch_latest_patch_versions("asdf"))


def env_run(env_path, command, *args, **kwargs):
    """
    Run a command on the context of the virtual environment at
    the specified path.
    """
    return subprocess.run([env_path / "bin" / command, *list(args)], **kwargs).returncode == 0  # noqa: S603


def env_check_output(env_path, command, *args):
    """
    Run a command on the context of the virtual environment at
    the specified path, and return the output.
    """
    return subprocess.check_output([env_path / "bin" / command, *list(args)]).decode("utf-8").strip()  # noqa: S603


def get_supported_versions(env_path):
    """
    Get ASDF Standard versions that are supported by the asdf library
    installed in the specified virtual environment.
    """
    script = r"""import asdf; print("\n".join(str(v) for v in asdf.versioning.supported_versions))"""
    output = env_check_output(env_path, "python3", "-c", script)
    return [asdf.versioning.AsdfVersion(v) for v in output.split("\n")]


def get_installed_version(env_path):
    """
    Get the version of the asdf library installed in the specified
    virtual environment.
    """
    script = r"""import asdf; print(asdf.__version__)"""
    return Version(env_check_output(env_path, "python3", "-c", script))


@pytest.fixture(scope="module", params=PATCH_VERSIONS, ids=[f"asdf=={v}" for v in PATCH_VERSIONS])
def asdf_version(request):
    """
    The (old) version of the asdf library under test.
    """
    return request.param


@pytest.fixture(scope="module")
def env_path(asdf_version, tmp_path_factory):
    """
    Path to the virtualenv where the (old) asdf library is installed.
    """
    path = tmp_path_factory.mktemp(f"asdf-{asdf_version}-env", numbered=False)

    assert subprocess.run(["virtualenv", str(path)]).returncode == 0  # noqa: S603,S607

    assert env_run(
        path,
        "pip",
        "install",
        f"asdf=={asdf_version}",
        capture_output=True,
    ), f"Failed to install asdf version {asdf_version}"

    return path


@pytest.fixture(autouse=True)
def _pushd_tmp_path(tmp_path):
    """
    Change the working directory, in case the user is running these
    tests from the repo root.  Python will import a module from the
    current working directory by preference, so this prevents us
    from accidentally comparing the current library code to itself.
    """
    original_cwd = os.getcwd()
    os.chdir(str(tmp_path))
    yield
    os.chdir(original_cwd)


@pytest.mark.remote_data()
def test_file_compatibility(asdf_version, env_path, tmp_path):
    # Sanity check to ensure we're not accidentally comparing
    # the current code to itself.
    installed_version = get_installed_version(env_path)
    assert installed_version == asdf_version, (
        f"The version of asdf in the virtualenv ({installed_version}) does "
        f"not match the version being tested ({asdf_version})"
    )

    # We can only test ASDF Standard versions that both library
    # versions support.
    current_supported_versions = set(asdf.versioning.supported_versions)
    old_supported_versions = set(get_supported_versions(env_path))
    standard_versions = list(current_supported_versions.intersection(old_supported_versions))

    if asdf_version < MIN_VERSION_1_0_0_STANDARD:
        standard_versions.remove(AsdfVersion("1.0.0"))

    # Confirm that this test isn't giving us a false sense of security.
    assert len(standard_versions) > 0

    for standard_version in sorted(standard_versions):
        # older versions of asdf reported support for 1.6.0 (the development
        # version of the standard) yet did not support the tags in that
        # version.
        if standard_version >= NDARRAY_1_1_0_STANDARD_VERSION:
            if asdf_version < NDARRAY_1_1_0_ASDF_VERSION:
                continue
        # Confirm that a file generated by the current version of the code
        # can be read by the older version of the library.
        if asdf_version >= MIN_VERSION_NEW_FILES:
            current_file_path = Path(str(tmp_path)) / "test-current.asdf"
            generate_file(current_file_path, standard_version)
            assert env_run(env_path, "python3", ASSERT_SCRIPT_PATH, current_file_path, capture_output=True), (
                f"asdf library version {asdf_version} failed to read an ASDF Standard {standard_version} "
                "file produced by this code"
            )

        # Confirm that a file generated by the older version of the library
        # can be read by the current version of the code.
        if asdf_version >= MIN_VERSION_OLD_FILES:
            old_file_path = Path(str(tmp_path)) / "test-old.asdf"
            assert env_run(
                env_path,
                "python3",
                GENERATE_SCRIPT_PATH,
                old_file_path,
                str(standard_version),
                capture_output=True,
            ), f"asdf library version {asdf_version} failed to generate an ASDF Standard {standard_version} file"
            assert_file_correct(old_file_path), (
                f"asdf library version {asdf_version} produced an ASDF Standard {standard_version}"
                "that this code failed to read"
            )



import datetime
import fractions
import warnings

import pytest

import asdf
from asdf.exceptions import AsdfWarning, ValidationError
from asdf.extension import Converter, Extension, ExtensionProxy
from asdf.tags.core import HistoryEntry
from asdf.testing import helpers


def test_history():
    ff = asdf.AsdfFile()
    assert "history" not in ff.tree
    ff.add_history_entry(
        "This happened",
        {"name": "my_tool", "homepage": "http://nowhere.org", "author": "John Doe", "version": "2.0"},
    )
    assert len(ff.tree["history"]["entries"]) == 1

    with pytest.raises(ValidationError, match=r".* is not valid under any of the given schemas"):
        ff.add_history_entry("That happened", {"author": "John Doe", "version": "2.0"})
    assert len(ff.tree["history"]["entries"]) == 1

    ff.add_history_entry("This other thing happened")
    assert len(ff.tree["history"]["entries"]) == 2

    assert isinstance(ff.tree["history"]["entries"][0]["time"], datetime.datetime)


def test_history_to_file(tmp_path):
    file_path = tmp_path / "history.asdf"

    with asdf.AsdfFile() as ff:
        ff.add_history_entry(
            "This happened",
            {"name": "my_tool", "homepage": "http://nowhere.org", "author": "John Doe", "version": "2.0"},
        )
        ff.write_to(file_path)

    with asdf.open(file_path) as ff:
        assert "entries" in ff.tree["history"]
        assert "extensions" in ff.tree["history"]
        assert len(ff.tree["history"]["entries"]) == 1

        entry = ff.tree["history"]["entries"][0]
        assert entry["description"] == "This happened"
        assert entry["software"]["name"] == "my_tool"
        assert entry["software"]["version"] == "2.0"

        # Test the history entry retrieval API
        entries = ff.get_history_entries()
        assert len(entries) == 1
        assert isinstance(entries, list)
        assert isinstance(entries[0], HistoryEntry)
        assert entries[0]["description"] == "This happened"
        assert entries[0]["software"]["name"] == "my_tool"


def test_old_history():
    """Make sure that old versions of the history format are still accepted"""

    yaml = """
history:
  - !core/history_entry-1.0.0
    description: "Here's a test of old history entries"
    software: !core/software-1.0.0
      name: foo
      version: 1.2.3
    """

    buff = helpers.yaml_to_asdf(yaml)
    with asdf.open(buff) as af:
        assert len(af.tree["history"]) == 1

        # Test the history entry retrieval API
        entries = af.get_history_entries()
        assert len(entries) == 1
        assert isinstance(entries, list)
        assert isinstance(entries[0], HistoryEntry)
        assert entries[0]["description"] == "Here's a test of old history entries"
        assert entries[0]["software"]["name"] == "foo"


def test_get_history_entries(tmp_path):
    """
    Test edge cases for the get_history_entries API. Other cases tested above
    """
    file_path = tmp_path / "empty.asdf"

    with asdf.AsdfFile() as af:
        af.write_to(file_path)

    # Make sure this works when there is no history section at all
    with asdf.open(file_path) as af:
        assert len(af["history"]["extensions"]) > 0
        assert len(af.get_history_entries()) == 0


def test_extension_metadata(tmp_path):
    file_path = tmp_path / "extension.asdf"

    ff = asdf.AsdfFile(version="1.6.0")
    ff.write_to(file_path)

    with asdf.open(file_path) as af:
        assert len(af.tree["history"]["extensions"]) == 1
        metadata = af.tree["history"]["extensions"][0]
        assert metadata.extension_uri == "asdf://asdf-format.org/core/extensions/core-1.6.0"
        assert metadata.extension_class == "asdf.extension._manifest.ManifestExtension"
        assert metadata.software["name"] == "asdf"
        assert metadata.software["version"] == asdf.__version__


def test_missing_extension_warning():
    yaml = """
history:
  extensions:
    - !core/extension_metadata-1.0.0
      extension_class: foo.bar.FooBar
      software: !core/software-1.0.0
        name: foo
        version: 1.2.3
    """

    buff = helpers.yaml_to_asdf(yaml)
    with pytest.warns(AsdfWarning, match=r"File was created with extension class 'foo.bar.FooBar'"), asdf.open(buff):
        pass


def test_extension_version_warning():
    uri = "asdf://somewhere.org/extensions/foo-1.0.0"
    package_name = "foo"
    file_package_version = "2.0.0"
    installed_package_version = "1.0.0"

    class FooExtension:
        extension_uri = uri

    yaml = f"""
history:
  extensions:
    - !core/extension_metadata-1.0.0
      extension_class: {FooExtension.__qualname__}
      extension_uri: {uri}
      software: !core/software-1.0.0
        name: {package_name}
        version: {file_package_version}
    """

    buff = helpers.yaml_to_asdf(yaml)
    with asdf.config_context() as cfg:
        cfg.add_extension(ExtensionProxy(FooExtension(), package_name, installed_package_version))
        with (
            pytest.warns(
                AsdfWarning,
                match=f"older package \\({package_name}=={installed_package_version}\\)",
            ),
            asdf.open(buff),
        ):
            pass

        buff.seek(0)

        # Make sure suppressing the warning works too
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            with asdf.open(buff, ignore_missing_extensions=True):
                pass


def test_strict_extension_check():
    yaml = """
history:
  extensions:
    - !core/extension_metadata-1.0.0
      extension_class: foo.bar.FooBar
      software: !core/software-1.0.0
        name: foo
        version: 1.2.3
    """

    buff = helpers.yaml_to_asdf(yaml)
    with (
        pytest.raises(
            RuntimeError,
            match=r"File was created with extension class .*, which is not currently installed",
        ),
        asdf.open(buff, strict_extension_check=True),
    ):
        pass

    # Make sure to test for incompatibility with ignore_missing_extensions
    buff.seek(0)
    with (
        pytest.raises(
            ValueError,
            match=r"'strict_extension_check' and 'ignore_missing_extensions' are incompatible options",
        ),
        asdf.open(buff, strict_extension_check=True, ignore_missing_extensions=True),
    ):
        pass


def test_metadata_with_custom_extension(tmp_path):
    class FractionConverter(Converter):
        tags = ["asdf://nowhere.org/tags/fraction-1.0.0"]
        types = [fractions.Fraction]

        def to_yaml_tree(self, obj, tag, ctx):
            return [obj.numerator, obj.denominator]

        def from_yaml_tree(self, node, tag, ctx):
            return fractions.Fraction(node[0], node[1])

    class FractionExtension(Extension):
        extension_uri = "asdf://nowhere.org/extensions/fraction-1.0.0"
        converters = [FractionConverter()]
        tags = FractionConverter.tags

    file_path = tmp_path / "custom_extension.asdf"

    with asdf.config_context() as config:
        config.add_extension(FractionExtension())

        tree = {"fraction": fractions.Fraction(2, 3)}

        with asdf.AsdfFile(tree) as ff:
            ff.write_to(file_path)

        # We expect metadata about both the Builtin extension and the custom one
        with asdf.open(file_path) as af:
            assert len(af["history"]["extensions"]) == 2

    with (
        pytest.warns(AsdfWarning, match=r"was created with extension"),
        asdf.open(
            file_path,
            ignore_unrecognized_tag=True,
        ),
    ):
        pass

    file_path_2 = tmp_path / "no_extension.asdf"

    # If we use the extension but we don't serialize any types that require it,
    # no metadata about this extension should be added to the file
    with asdf.config_context() as config:
        config.add_extension(FractionExtension())

        tree2 = {"x": list(range(10))}
        with asdf.AsdfFile(tree2) as ff:
            ff.write_to(file_path_2)

        with asdf.open(file_path_2) as af:
            assert len(af["history"]["extensions"]) == 1

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        with asdf.open(file_path_2):
            pass

    file_path_3 = tmp_path / "custom_extension2.asdf"

    with asdf.config_context() as config:
        config.add_extension(FractionExtension())
        # Make sure that this works even when constructing the tree on-the-fly
        with asdf.AsdfFile() as ff:
            ff.tree["fraction"] = fractions.Fraction(4, 5)
            ff.write_to(file_path_3)

        with asdf.open(file_path_3) as af:
            assert len(af["history"]["extensions"]) == 2
